unit calibration;

interface

uses
  Contnrs,
  donnee, stereo_bitan,// from Convertisseur
  reference, carte, compute;


const
  Dist_Limit = 3;
  
type

  TMethod = record
    Method : integer;
    UnknownProjection : boolean;
    Projection : integer;
    Nord, Isotrop, knownScale : boolean;
    Scale : real;
  end;

  TCalibrationProject = class
  private
    FMap: TImageCarte;
    ListRefPoint : TObjectList;
    FChanged: boolean;
    FCurrentFileName: string;
    FCalculated: boolean;
    FMethod: TMethod;
    LocalProjection : TStereo_bitan;
    Cal_Parameters : TCal_Parameters;
    function GetRefPoint(index: integer): TReferencePoint;
    procedure SetMap(const Value: TImageCarte);
    procedure SetChanged(const Value: boolean);
    procedure SetCurrentFileName(const Value: string);
    function GetProposedFileName: string;
    function GetNbRefPoint: integer;
    procedure SetCalculated(const Value: boolean);
    procedure SetMethod(const Value: TMethod);
    procedure Point2WGS84(Num_Point: integer; var Lat, Long : real); // convert the point from his own coordinate system to WGS84
  public
//    Construction : TConstruction;
    property RefPoint[index : integer] : TReferencePoint read GetRefPoint;
    property NbRefPoint : integer read GetNbRefPoint;
    procedure AddRefPoint(NewPoint : TReferencePoint);
    procedure DeleteRefPoint(index : integer);
    property Map : TImageCarte read FMap write SetMap;
    property Changed : boolean read FChanged write SetChanged;
    property CurrentFileName : string read FCurrentFileName write SetCurrentFileName;
    property ProposedFileName : string read GetProposedFileName;
    property Calculated : boolean read FCalculated write SetCalculated;
    property Method : TMethod read FMethod write SetMethod;
    constructor Create; virtual;
    destructor Destroy; override;
    procedure LoadFromFile(FileName : string);
    procedure SaveToFile(FileName : string);
    procedure Calculate; // determine the calibration parameters
    procedure Pixel2WGS84(X, Y : integer; var Lat, Long : real);
    procedure WGS842Pixel(Lat, Long : real; var X, Y : real);
    function Error_Pixel(Num_Point : integer) : real; // give the error at Num_Point in pixel
    function Error_Metric(Num_Point : integer) : real; // give the error at Num_Point in meter
    procedure Scales(var scale1, scale2, rotation : real); // give the scale with 2 values if anysotropic and the overall rotation
    procedure Extrapole4pt(FileName : string);
    procedure ExportOzi(FileName : string);
    procedure ExportPicLayer(FileName : string);
  end;

implementation

{ TCalibrationProject }

uses
  Classes, SHFolder, SysUtils, Math,
  ancetre, datum, ellipsoide, outils; // from Convertisseur

procedure TCalibrationProject.AddRefPoint(NewPoint: TReferencePoint);
begin
  ListRefPoint.Add(NewPoint);
  Changed:=true;
end;

constructor TCalibrationProject.Create;
const
  Taille=256;
var
  Chemin : string;
  resultat : array[0..taille] of char;
  Flux : TFileStream;
begin
  ListRefPoint := TObjectList.Create;

  Donnees:=TDonnees.Create;
  // trying to get datas from Convertisseur, if any
  SHGetFolderPath(0,CSIDL_APPDATA,0,taille,resultat);
  Chemin:=StrPas(resultat)+'\convertisseur\donnees.txt';
  try
    Flux:=TFileStream.Create(Chemin,fmOpenRead+fmShareDenyWrite);
    Donnees.LoadFromStream(Flux);
    Flux.Free;
  except
  end;

  with FMethod do
  begin
    Method :=0;
    UnknownProjection := true;
    Projection := 0;
    Nord :=false;
    Isotrop := false;
    knownScale := false;
    Scale := 0;
  end;
  Calculated:=false;

//  Construction := TConstruction.Create;

  Changed:=false;
end;

procedure TCalibrationProject.DeleteRefPoint(index: integer);
begin
  assert(index in [0..ListRefPoint.Count-1],'Index out of limits in TCalibrationProject.DeleteRefPoint');
  ListRefPoint.Delete(index);
  Changed:=true;
end;

destructor TCalibrationProject.Destroy;
begin
  ListRefPoint.Free;
  Donnees.Free;

//  Construction.Free;

  if FMap<>nil then
    FMap.Free;
  inherited;
end;

function TCalibrationProject.GetNbRefPoint: integer;
begin
  result:=ListRefPoint.Count;
end;

function TCalibrationProject.GetProposedFileName: string;
begin
  if FMap<>nil then
    result:=ChangeFileExt(FMap.SourceFile,'.txt')
  else
    result:='';
end;

function TCalibrationProject.GetRefPoint(index: integer): TReferencePoint;
begin
  assert(index in [0..ListRefPoint.Count-1],'Index out of limits in TCalibrationProject.GetRefPoint');
  result:=ListRefPoint[index] as TReferencePoint;
end;

procedure TCalibrationProject.LoadFromFile(FileName: string);
var
  F : textFile;
  S : string;
  I : integer;
  MemoryStream : TMemoryStream;
  NewRefPoint : TReferencePoint;
begin
  FCurrentFileName := FileName;
  assignFile(F,FCurrentFileName);
  reset(F);
  try

// map file
    readln(F,S);
    if UpperCase(S)<>'[MAP]' then
      Abort;
    readln(F,S);
    if S<>'' then
    begin
//      S:=ExpandFileName(S);
      S:=ExpandFileName(ExtractFilePath(FileName)+S);
      if Map<>nil then
        Map.Free;
      Map:=TImageCarteBasic.Create(S);
    end;

// list of ellipsoid, datum and projection used in RefPoint and Method
    readln(F,S);
    if UpperCase(S)<>'[GEODESY]' then
      Abort;

    MemoryStream:=TMemoryStream.Create;
    try

      readln(F,S);
      while S[1]<>'[' do
      begin
        MemoryStream.Clear;
        MemoryStream.Write(S[1],length(S));
        MemorySTream.Position:=0;
        try
          Donnees.LoadLigne(MemoryStream);
        except // exception can be raised if the geodesic data is already loaded by default
               // just trap
        end;
        readln(F,S);
      end;
    finally
      MemoryStream.Free;
    end;

// list of reference points
    if UpperCase(S)<>'[POINTS]' then
      Abort;
    readln(F,S);
    while S[1]<>'[' do
    begin
      NewRefPoint := TReferencePoint.Create;
      NewRefPoint.LoadFromString(S);
      AddRefPoint(NewRefPoint);
      readln(F,S);
    end;

// calibration method
    if UpperCase(S)<>'[METHOD]' then
      Abort;
    readln(F,S);
    FMethod.Method:=StrToInt(S);
    if Method.Method=0 then
    begin // more parameters for affine method
      readln(F,S);
      FMethod.UnknownProjection:=UpperCase(S)='UNKNOWN';
      if not FMethod.UnknownProjection then
      begin
        for I:=0 to Donnees.NbProjection-1 do
          if (Donnees.LesProjection[I] as TAncetre).Nom=S then
            FMethod.Projection:=I;
        readln(F,S);
        FMethod.Nord:=UpperCase(S)='TRUE';
      end;
      readln(F,S);
      FMethod.Isotrop:=UpperCase(S)='TRUE';
      readln(F,S);
      FMethod.KnownScale:=UpperCase(S)='TRUE';
      if FMethod.KnownScale then
      begin
        readln(F,S);
        FMethod.Scale:=StrToFloat(S);
      end;
    end;

    // construction
{    if not Eof(F) then
    begin
      readln(F, S);
      if UpperCase(S)<>'[CONSTRUCTION]' then
        Abort;
      Construction.LoadFromFile(F);
    end;
 }

  finally
    CloseFile(F);
    FChanged:=false;
    Calculated:=false;
  end;
end;

procedure TCalibrationProject.SaveToFile(FileName: string);
var
  Flux : TFileStream;
  UsedEllipsoide, UsedDatum, UsedProjection, UsedMultiple : array of boolean;
  I, Which : integer;
  S : string;
begin
  FCurrentFileName := FileName;
  Flux := TFileStream.Create(CurrentFileName, fmCreate);
  try
// map file
    S:='[map]'+#13#10;
    Flux.Write(S[1], length(S));
    if Map.SourceFile='' then
      S:=#13#10
    else
      S:=ExtractRelativePath(CurrentFileName, Map.SourceFile)+#13#10;
    Flux.Write(S[1], length(S));

// list of ellipsoid, datum and projection used in RefPoint and Method

    setLength(UsedEllipsoide, Donnees.NbEllipsoide);
    for I:=low(UsedEllipsoide) to high(UsedEllipsoide) do
      UsedEllipsoide[I]:=false;
    setLength(UsedDatum, Donnees.NbDatum);
    for I:=low(UsedDatum) to high(UsedDatum) do
      UsedDatum[I]:=false;
    setLength(UsedProjection, Donnees.NbProjection);
    for I:=low(UsedProjection) to high(UsedProjection) do
      UsedProjection[I]:=false;
    setLength(UsedMultiple, Donnees.NbMultiple);
    for I:=low(UsedMultiple) to high(UsedMultiple) do
      UsedMultiple[I]:=false;

    // looking for datum and projections used in RefPoints
    for I:=0 to NbRefPoint-1 do
    begin
      if RefPoint[I].Num_Proj<Donnees.NbDatum then
        UsedDatum[RefPoint[I].Num_Proj]:=true
      else
      begin
        if RefPoint[I].Num_Proj<Donnees.NbDatum+Donnees.NbProjection then
        begin
          Which:=RefPoint[I].Num_Proj-Donnees.NbDatum;
          UsedProjection[Which]:=true;
          UsedDatum[Donnees.NumeroDatum((Donnees.LesProjection[Which] as TAncetre).Datum.Nom)]:=true;
        end
        else
        begin
          Which:=RefPoint[I].Num_Proj-Donnees.NbDatum-Donnees.NbProjection;
          UsedMultiple[Which]:=true;
          UsedDatum[Donnees.NumeroDatum((Donnees.LesMultiple[Which] as TAncetre).Datum.Nom)]:=true;
        end;
      end;
    end;
    // adding the projection used by the calibration method
    if not Method.UnknownProjection then
    begin
      UsedProjection[Method.Projection]:=true;
      UsedDatum[Donnees.NumeroDatum((Donnees.LesProjection[Method.Projection] as TAncetre).Datum.Nom)]:=true;
    end;
    // looking for ellipsoid used by the used datum
    for I:=0 to Donnees.NbDatum-1 do
      if UsedDatum[I] then
        UsedEllipsoide[Donnees.NumeroEllipsoide((Donnees.LesDatum[I] as TDatum).Ellipsoide.Nom)]:=true;
    // storing everything used
    S:='[geodesy]'+#13#10;
    Flux.Write(S[1], length(S));
    for I:=1 to Donnees.NbEllipsoide-1 do // skip first => GRS80
      if UsedEllipsoide[I] then
        (Donnees.LesEllipsoide[I] as TEllipsoide).SaveToStream(Flux);
    for I:=1 to Donnees.NbDatum-1 do // skip first => WGS84
      if UsedDatum[I] then
        (Donnees.LesDatum[I] as TDatum).SaveToStream(Flux);
    for I:=0 to Donnees.NbProjection-1 do
      if UsedProjection[I] then
        (Donnees.LesProjection[I] as TAncetre).SaveToStream(Flux);
    for I:=0 to Donnees.NbMultiple-1 do
      if UsedMultiple[I] then
        (Donnees.LesMultiple[I] as TAncetre).SaveToStream(Flux);

// list of reference points
    S:='[points]'+#13#10;
    Flux.Write(S[1], length(S));
    for I:=0 to NbRefPoint-1 do
    begin
      S:=(RefPoint[I] as TReferencePoint).SaveToString+#13#10;
      Flux.Write(S[1], length(S));
    end;

// calibration method
    S:='[method]'+#13#10;
    Flux.Write(S[1], length(S));
    with Method do
    begin
      S:=IntToStr(Method)+#13#10;
      Flux.Write(S[1], length(S));
      if Method=0 then
      begin // more parameters for affine method
        if UnknownProjection then
        begin
          S:='Unknown'+#13#10;
          Flux.Write(S[1], length(S));
        end
        else
        begin
          S:=(Donnees.LesProjection[Projection] as TAncetre).Nom+#13#10;
          Flux.Write(S[1], length(S));
          if Nord then
            S:='TRUE'+#13#10
          else
            S:='FALSE'+#13#10;
          Flux.Write(S[1], length(S));
        end;
        if Isotrop then
          S:='TRUE'+#13#10
        else
          S:='FALSE'+#13#10;
        Flux.Write(S[1], length(S));
        if knownScale then
        begin
          S:='TRUE'+#13#10;
          Flux.Write(S[1], length(S));
          S:=FloatToStrF(Scale,ffFixed,10,3)+#13#10;
          Flux.Write(S[1], length(S));
        end
        else
        begin
          S:='FALSE'+#13#10;
          Flux.Write(S[1], length(S));
        end;
      end;
    end;

// construction
{    S:='[construction]'+#13#10;
    Flux.Write(S[1], length(S));
    Construction.SaveToFile(Flux);
 }
    Changed:=false;
  finally
    Flux.Free;
  end;
end;

procedure TCalibrationProject.SetChanged(const Value: boolean);
begin
  FChanged := Value;
  if Value then
    Calculated:=false;
end;

procedure TCalibrationProject.SetCurrentFileName(const Value: string);
begin
  FCurrentFileName := Value;
end;

procedure TCalibrationProject.SetMap(const Value: TImageCarte);
begin
  if FMap<>nil then
    FMap.Free;
  FMap := Value;
  Changed:=true;
end;

procedure TCalibrationProject.SetMethod(const Value: TMethod);
begin
  FMethod := Value;
  Changed:=true;
end;

procedure TCalibrationProject.SetCalculated(const Value: boolean);
begin
  FCalculated := Value;
  if (not Value) and (LocalProjection<>nil) then
  begin
    LocalProjection.Free;
    LocalProjection:=nil;
  end;
end;

procedure TCalibrationProject.Calculate; // determine the calibration parameters
var
  Xcart, Ycart, Zcart, Xm, Ym, Zm : real;
  I, IndiceSource : integer;
  Lat, Long, SProj : string;
  Lati, Longi : real;
  Relations : TRelations;
begin
  assert((Method.Method=0) or Method.UnknownProjection,'Projection must not be defined when using Homographic or Quadradic in TCalibrationProject.Calculate');
  if not (Method.Method=0) or Method.UnknownProjection then
  // we build a local projection by defining a stereographic projection with
  // origin point at the average position of the reference points
  begin
    if LocalProjection<>nil then
      LocalProjection.Free;
    Xm:=0; Ym:=0; Zm:=0;
    for I:=0 to ListRefPoint.Count-1 do
     with RefPoint[I] do
    begin
      case ProjectionType of
        ptDatum :
          if Paris then
            IndiceSource:=1
          else
            IndiceSource:=0;
        ptMultiple : IndiceSource:=Zone;
      else
        IndiceSource:=0;
      end;

      donnees.VersGeocentrique(Name_Proj, Latitude, Longitude, IndiceSource, Xcart, Ycart, Zcart);
      Xm:=Xm+Xcart;
      Ym:=Ym+Ycart;
      Zm:=Zm+Zcart;
    end;
    Xm:=Xm/NbRefPoint;
    Ym:=Ym/NbRefPoint;
    Zm:=Zm/NbRefPoint;
    IndiceSource:=0;
    donnees.DepuisGeocentrique(Xm, Ym, Zm, 'WGS84', IndiceSource, Lat, Long, uaDegre);
    LocalProjection := TStereo_bitan.Create;
    LocalProjection.Datum:=(donnees.LesDatum[0] as TDatum); // WGS84 datum
    LocalProjection.ParamReal[0]:=1; // scale factor at the origine
    LocalProjection.ParamReal[1]:=0; // false Easting
    LocalProjection.ParamReal[2]:=0; // false Northing
    LocalProjection.ParamAngle[0]:=Long; // origin point
    LocalProjection.ParamAngle[1]:=Lat; // origin point
    LocalProjection.Valid:=true;
  end;

  // converting the reference points from their own system to the map projection
  setlength(Relations,ListRefPoint.Count);
  for I:=0 to ListRefPoint.Count-1 do
   with RefPoint[I] do
  begin
    Relations[I].PixelX:=XPixel;
    Relations[I].PixelY:=YPixel;

    case ProjectionType of
      ptDatum :
        if Paris then
          IndiceSource:=1
        else
          IndiceSource:=0;
      ptMultiple : IndiceSource:=Zone;
    else
      IndiceSource:=0;
    end;
    donnees.VersGeocentrique(Name_Proj, Latitude, Longitude, IndiceSource, Xcart, Ycart, Zcart);

    if Method.UnknownProjection then
    begin
      LocalProjection.Datum.VersLatLong(Xcart, Ycart, Zcart, false, Lati, Longi);
      LocalProjection.VersGrille(Lati, Longi, Relations[I].CoordX, Relations[I].CoordY);
    end
    else
    begin
      SProj:=(Donnees.LesProjection[Method.Projection] as TAncetre).Nom;
      donnees.DepuisGeocentriqueNum(Xcart, Ycart, Zcart, SProj,
             IndiceSource, Relations[I].CoordX, Relations[I].CoordY);
    end;
  end;

  // calculating the calibration itself
  case Method.Method of
    0 : begin // affine
      Setlength(Cal_Parameters[0],6);
      Setlength(Cal_Parameters[1],6);
      if Method.Isotrop then
      begin
        if Method.Nord then
        begin
          if Method.knownScale then
            Compute_Affine_Isotrop_North_Scale(Relations , Method.Scale, Cal_Parameters)
          else
            Compute_Affine_Isotrop_North(Relations, Cal_Parameters);
        end
        else
        begin
          if Method.knownScale then
            Compute_Affine_Isotrop_Scale(Relations , Method.Scale, Cal_Parameters)
          else
            Compute_Affine_Isotrop(Relations, Cal_Parameters);
        end;
      end
      else
        Compute_Affine_svd(Relations, Cal_Parameters);
    end;
    1 : begin // homographic
      Setlength(Cal_Parameters[0],8);
      Setlength(Cal_Parameters[1],8);
      Compute_Homographic(Relations, Cal_Parameters);
    end;
    2 : begin // quadratic
      Setlength(Cal_Parameters[0],12);
      Setlength(Cal_Parameters[1],12);
      Compute_Quadratic(Relations, Cal_Parameters);
    end;
  end;
  Calculated:=true;

  // calculating the error of each point
  for I:=0 to ListRefPoint.Count-1 do
    RefPoint[I].ErrorDistance:=Error_Pixel(I);
end;

procedure TCalibrationProject.Pixel2WGS84(X, Y : integer; var Lat, Long : real);
var
  U, V, Latitude, Longitude : real;
  Xcart, Ycart, Zcart : real;
  IndiceSource : integer;
  SProj : string;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.Pixel2WGS84');
  // converting from pixels to local projection
  case Method.Method of
    0 : begin // affine
      assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in  TCalibrationProject.Pixel2WGS84');
      U:=X*Cal_Parameters[0, 0]+Y*Cal_Parameters[0, 1]+Cal_Parameters[0, 2];
      V:=X*Cal_Parameters[0, 3]+Y*Cal_Parameters[0, 4]+Cal_Parameters[0, 5];
    end;
    1 : begin // homographic
      assert(length(Cal_Parameters[0])=8,'Cal_Parameters not set to the right length in  TCalibrationProject.Pixel2WGS84');

      U:=(X * Cal_Parameters[0, 0] + Y * Cal_Parameters[0, 1] + Cal_Parameters[0, 2]) /
                  (X * Cal_Parameters[0, 6] + Y * Cal_Parameters[0, 7] + 1.0);
      V:=(X * Cal_Parameters[0, 3] + Y * Cal_Parameters[0, 4] + Cal_Parameters[0, 5]) /
                  (X * Cal_Parameters[0, 6] + Y * Cal_Parameters[0, 7] + 1.0);

    end;
    2 : begin // quadratic
      assert(length(Cal_Parameters[0])=12,'Cal_Parameters not set to the right length in  TCalibrationProject.Pixel2WGS84');

      U:= Cal_Parameters[0, 0] * sqr(X) + Cal_Parameters[0, 1] * sqr(Y) + Cal_Parameters[0, 2] * X * Y +
          Cal_Parameters[0, 3] * X + Cal_Parameters[0, 4] * Y + Cal_Parameters[0, 5];
      V:= Cal_Parameters[0, 6] * sqr(X) + Cal_Parameters[0, 7] * sqr(Y) + Cal_Parameters[0, 8] * X * Y +
          Cal_Parameters[0, 9] * X + Cal_Parameters[0, 10] * Y + Cal_Parameters[0, 11];

    end;
  end;
  // converting from local projection to Lat, Long (WGS84)
  if Method.UnknownProjection then
  begin
    assert(LocalProjection<>nil,'Local projection not calculated in TCalibrationProject.Pixel2WGS84');
    LocalProjection.VersLatLong(U, V, Latitude, Longitude);
    LocalProjection.Datum.VersCartesien(Latitude, Longitude, false, Xcart, Ycart, Zcart);
  end
  else
  begin
    IndiceSource:=0;
    SProj:=(Donnees.LesProjection[Method.Projection] as TAncetre).Nom;
    donnees.VersGeocentriqueNum(SProj, U, V, IndiceSource, XCart, YCart, Zcart);
  end;
  IndiceSource:=0;
  donnees.DepuisGeocentriqueNum(Xcart, YCart, Zcart,'WGS84',IndiceSource,Lat,Long);

end;

procedure TCalibrationProject.WGS842Pixel(Lat, Long : real; var X, Y : real);
var
  U, V, Latitude, Longitude : real;
{  InvDet, g, h, i, j, k, l : real;
  a, b, c, d, e, f : real; }
  Xcart, Ycart, Zcart : real;
  IndiceDest : integer;
  SProj : string;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.WGS842Pixel');
  // converting from Lat, Long (WGS84) to local projection
  donnees.VersGeocentriqueNum('WGS84',Lat, Long,0,Xcart,Ycart, Zcart);
  if Method.UnknownProjection then
  begin
    assert(LocalProjection<>nil,'Local projection not calculated in TCalibrationProject.WGS842Pixel');
    LocalProjection.Datum.VersLatLong(Xcart, Ycart, Zcart, false, Latitude, Longitude);
    LocalProjection.VersGrille(Latitude, Longitude, U, V);
  end
  else
  begin
    SProj:=(Donnees.LesProjection[Method.Projection] as TAncetre).Nom;
    donnees.DepuisGeocentriqueNum(XCart, YCart, ZCart,SProj,IndiceDest, U, V);
  end;

  // converting from local projection to pixels
  case Method.Method of
    0 : begin // affine
      assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in  TCalibrationProject.WGS842Pixel');

      X := Cal_Parameters[1, 0] * U + Cal_Parameters[1, 1] * V + Cal_Parameters[1, 2];
      Y := Cal_Parameters[1, 3] * U + Cal_Parameters[1, 4] * V + Cal_Parameters[1, 5];
    end;
    1 : begin // homographic
      assert(length(Cal_Parameters[1])=8,'Cal_Parameters not set to the right length in  TCalibrationProject.WGS842Pixel');

      X := (Cal_Parameters[1, 0] * U + Cal_Parameters[1, 1] * V + Cal_Parameters[1, 2]) / (Cal_Parameters[1, 6] * U + Cal_Parameters[1, 7] * V + 1);
      Y := (Cal_Parameters[1, 3] * U + Cal_Parameters[1, 4] * V + Cal_Parameters[1, 5]) / (Cal_Parameters[1, 6] * U + Cal_Parameters[1, 7] * V + 1);

    end;
    2 : begin // quadratic
      assert(length(Cal_Parameters[1])=12,'Cal_Parameters not set to the right length in  TCalibrationProject.WGS842Pixel');

      X:= Cal_Parameters[1, 0] * sqr(U) + Cal_Parameters[1, 1] * sqr(V) + Cal_Parameters[1, 2] * U * V +
          Cal_Parameters[1, 3] * U + Cal_Parameters[1, 4] * V + Cal_Parameters[1, 5];
      Y:= Cal_Parameters[1, 6] * sqr(U) + Cal_Parameters[1, 7] * sqr(V) + Cal_Parameters[1, 8] * U * V +
          Cal_Parameters[1, 9] * U + Cal_Parameters[1, 10] * V + Cal_Parameters[1, 11];

    end;
  end;
end;

procedure TCalibrationProject.Point2WGS84(Num_Point: integer; var Lat, Long : real);
// convert the point from his own coordinate system to WGS84
var
  Xcart, Ycart, Zcart : real;
  IndiceSOurce : integer;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.Point2WGS84');
  assert(Num_Point in [0..NbRefPoint-1],'The point asked in TCalibrationProject.Point2WGS84 is out of the limits.');

  with RefPoint[Num_Point] do
  begin
    case ProjectionType of
      ptDatum :
        if Paris then
          IndiceSource:=1
        else
          IndiceSource:=0;
      ptMultiple : IndiceSource:=Zone;
    else
      IndiceSource:=0;
    end;
    donnees.VersGeocentrique(Name_Proj, Latitude, Longitude, IndiceSource, Xcart, Ycart, Zcart);
  end;
  IndiceSource := 0;
  donnees.DepuisGeocentriqueNum(Xcart, Ycart, Zcart, 'WGS84', IndiceSource, Lat, Long);
end;

function TCalibrationProject.Error_Pixel(Num_Point : integer) : real; // give the error at Num_Point in pixel
var
  Latitude, Longitude : real;
  XPix, YPix : real;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.Error_Pixel');
  assert(Num_Point in [0..NbRefPoint-1],'The point asked in TCalibrationProject.Error_Pixel is out of the limits.');
  Point2WGS84(Num_Point,Latitude, Longitude);
  WGS842Pixel(Latitude, Longitude, XPix, YPix);
  result:=sqrt(sqr(XPix-RefPoint[Num_Point].XPixel)+
                 sqr(YPix-RefPoint[Num_Point].YPixel));

end;

function TCalibrationProject.Error_Metric(Num_Point : integer) : real; // give the error at Num_Point in meter
const
  R : real =6374892.5; // earth radius
var
  Latitude, Longitude : real;
  Pixel_Lat, Pixel_Long : real;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.Error_Metric');
  assert(Num_Point in [0..NbRefPoint-1],'The point asked in TCalibrationProject.Error_Metric is out of the limits.');
  Point2WGS84(Num_Point,Latitude, Longitude);
  Pixel2WGS84(RefPoint[Num_Point].XPixel,
              RefPoint[Num_Point].YPixel,
              Pixel_Lat, Pixel_Long);
  result:=R*arccos(sin(Latitude)*sin(Pixel_Lat)+cos(Latitude)*cos(Pixel_Lat)*cos(Pixel_Long-Longitude)); // calculating distance on the sphere
end;

procedure TCalibrationProject.Scales(var scale1, scale2, rotation : real);
// give the scale with 2 values if anysotropic and the overall rotation
// calculate it in the center of the map for homographic and quadratic

  procedure Scale_Affine(a, b, d, e : real);
  // caculate the scales in the affine case
  // U := a*x + b*y (+ c);
  // V := d*x + e*y (+ f);
  // with a unit vector input (cos(Theta); sin(Theta)), we are looking for local min/max of R = (U + V)^0.5
  // i.e. where d(R)/dTheta = 0
  // for homographic: only valide at the center where X=0 and Y=0 in the local projection
  var
    Theta, Vert, Hori : real;
  begin
    if IsZero(E) then
      rotation := Pi/2
    else
      rotation := arctan(b/e);

    Hori := sqr(a) + sqr(e) - sqr(b) - sqr(d);
    Vert := a * b + d * e;
    if IsZero(Hori) then
      Theta := Pi/4 // map is turned exactly of 45
    else
      Theta := 0.5 * arctan(2*Vert/Hori);
    Scale1 := sqrt (sqr(a*cos(Theta)+b*sin(Theta)) + sqr(d*cos(Theta)+e*sin(Theta)));
    Theta := Theta + Pi/2;
    Scale2 := sqrt (sqr(a*cos(Theta)+b*sin(Theta)) + sqr(d*cos(Theta)+e*sin(Theta)));
  end;

var
  RatioX, RatioY : real;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.Scales');
  case Method.Method of
    0: begin // affine
      assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in TCalibrationProject.Scales');
      Scale_Affine(Cal_Parameters[0, 0], Cal_Parameters[0, 1],
                   Cal_Parameters[0, 3], Cal_Parameters[0, 4]);
    end;
    1: begin // homographic
      assert(length(Cal_Parameters[0])=8,'Cal_Parameters not set to the right length in TCalibrationProject.Scales');
      // we do a linear approximation in the center of the image and we go back to affine case
      RatioX := (Cal_Parameters[0, 0] * Map.Width/2 + Cal_Parameters[0, 1] * Map.Height/2 + Cal_Parameters[0, 2])/
                (Cal_Parameters[0, 6] * Map.Width/2 + Cal_Parameters[0, 7] * Map.Height/2 + 1);
      RatioY := (Cal_Parameters[0, 3] * Map.Width/2 + Cal_Parameters[0, 4] * Map.Height/2 + Cal_Parameters[0, 5])/
                (Cal_Parameters[0, 6] * Map.Width/2 + Cal_Parameters[0, 7] * Map.Height/2 + 1);
      Scale_Affine(Cal_Parameters[0, 0] - Cal_Parameters[0, 6] / RatioX,
                   Cal_Parameters[0, 1] - Cal_Parameters[0, 7] / RatioX,
                   Cal_Parameters[0, 3] - Cal_Parameters[0, 6] / RatioY,
                   Cal_Parameters[0, 4] - Cal_Parameters[0, 7] / RatioY);
    end;
    2 : begin // quadratic
      assert(length(Cal_Parameters[0])=12,'Cal_Parameters not set to the right length in TCalibrationProject.Scales');
      // we do a linear approximation in the center of the image and we go back to affine case
      Scale_Affine(Cal_Parameters[0, 0] * Map.Width + Cal_Parameters[0, 2] * Map.Height/2 + Cal_Parameters[0, 3],
                   Cal_Parameters[0, 1] * Map.Height + Cal_Parameters[0, 2] * Map.Width/2 + Cal_Parameters[0, 4],
                   Cal_Parameters[0, 6] * Map.Width + Cal_Parameters[0, 8] * Map.Height/2 + Cal_Parameters[0, 9],
                   Cal_Parameters[0, 7] * Map.Height + Cal_Parameters[0, 8] * Map.Width/2 + Cal_Parameters[0, 10]);
    end;
  end;
end;

procedure TCalibrationProject.ExportOzi(FileName : string);
// export the calibration using all points
// * OziExplorer Map Data File format is used

  function LatToStr(Lat : real) : string;
  begin
    if Lat<0 then
    begin
      Lat:=-Lat;
      result:='S';
    end
    else
      Result:='N';
    result:=format('%.2d, %.2d.%.4d,',[trunc(Lat),trunc(frac(Lat)*60),trunc(frac(Lat*60)*10000)])+Result;
  end;
  function LongToStr(Long : real) : string;
  begin
    if Long<0 then
    begin
      Long:=-Long;
      result:='W';
    end
    else
      Result:='E';
    result:=format('%.3d, %.2d.%.4d,',[trunc(Long),trunc(frac(Long)*60),trunc(frac(Long*60)*10000)])+Result;
  end;

var
  I : integer;
  Lat, Long, Scale : real;
  F : TextFile;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.ExportOzi');

  AssignFile(F,FileName);
  rewrite(F);
  try
    writeln(F,'OziExplorer Map Data File Version 2.2');
    writeln(F,ExtractFileName(FileName));
    writeln(F,FileName);
    writeln(F,'1, Map Code,');
    writeln(F,'WGS 84,WGS 84,   0.0000,   0.0000,WGS 84');
    writeln(F,'Reserved 1');
    writeln(F,'Reserved 2');
    writeln(F,'Magnetic Variation,,,E');
    writeln(F,'Map Projection,Mercator,PolyCal,No,AutoCalOnly,No,BSBUseWPX,No');

    for I:=0 to NbRefPoint-1 do
     with RefPoint[I] do
    begin
      Point2WGS84(I, Lat, Long);
      Lat:=Lat/Pi*180; Long:=Long/Pi*180;
      writeln(F,Format('Point%.2d,xy, %.4d, %.4d,in, deg,  %s, %s, grid,   ,           ,           ,N',
                                            [I+1, XPixel, YPixel, LatToStr(Lat), LongToStr(Long)]));
    end;

    for I:=NbRefPoint+1 to 9 do
      writeln(F,Format('Point%.2d,xy,     ,     ,in, deg,    ,        ,N,    ,        ,E, grid,   ,           ,           ,N',[I]));

    for I:=max(10, NbRefPoint+1) to 30 do
      writeln(F,Format('Point%d,xy,     ,     ,in, deg,    ,        ,,    ,        ,, grid,   ,           ,           ,',[I]));

    writeln(F,'Map Feature = MF ; Map Comment = MC     These follow if they exist');
    writeln(F,'Track File = TF      These follow if they exist');
    writeln(F,'Moving Map Parameters = MM?    These follow if they exist');
    writeln(F,'MM0,Yes');
    writeln(F,'MMPNUM,4');
    writeln(F,'MMPXY,1,0,0');
    writeln(F,Format('MMPXY,2,%d,0',[Map.Width]));
    writeln(F,Format('MMPXY,3,%d,%d',[Map.Width, Map.Height]));
    writeln(F,Format('MMPXY,4,0,%d',[Map.Height]));
    Pixel2WGS84(0, 0, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,1,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Pixel2WGS84(Map.Width, 0, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,2,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Pixel2WGS84(Map.Width, Map.Height, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,3,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Pixel2WGS84(0, Map.Height, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,4,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Scale:=0.5*(sqrt(sqr(Cal_Parameters[0, 0])+sqr(Cal_Parameters[0, 1]))+sqrt(sqr(Cal_Parameters[0, 3])+sqr(Cal_Parameters[0, 4])));
    writeln(F,Format('MM1B,%d.%.6d',[trunc(Scale),trunc(frac(Scale)*1000000)]));
    writeln(F,'MOP,Map Open Position,0,0');
    writeln(F,Format('IWH,Map Image Width/Height,%d,%d',[Map.Width, Map.Height]));

  finally
    CloseFile(F);
  end;
end;

procedure TCalibrationProject.Extrapole4pt(FileName : string);
// export the calibration using for 4 points for
// * OziExplorer Map Data File format is used
// * for TritonRMP (*.gmap)

  function LatToStr(Lat : real) : string;
  begin
    if Lat<0 then
    begin
      Lat:=-Lat;
      result:='S';
    end
    else
      Result:='N';
    result:=format('%.2d, %.2d.%.4d,',[trunc(Lat),trunc(frac(Lat)*60),trunc(frac(Lat*60)*10000)])+Result;
  end;
  function LongToStr(Long : real) : string;
  begin
    if Long<0 then
    begin
      Long:=-Long;
      result:='W';
    end
    else
      Result:='E';
    result:=format('%.3d, %.2d.%.4d,',[trunc(Long),trunc(frac(Long)*60),trunc(frac(Long*60)*10000)])+Result;
  end;

const
  Beta : real = 0.577; // 1/sqrt(3)
var
  I : integer;
  XMin, XMax, YMin, YMax : integer;
  Lat, Long, Scale : real;
  F : TextFile;
  OldDecimalSeparator : char;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.Extrapole4pt');

// export the calibration using for 4 points
// using of OziExplorer Map Data File format

  AssignFile(F,FileName);
  rewrite(F);
  try
    writeln(F,'OziExplorer Map Data File Version 2.2');
    writeln(F,ExtractFileName(FileName));
    writeln(F,FileName);
    writeln(F,'1, Map Code,');
    writeln(F,'WGS 84,WGS 84,   0.0000,   0.0000,WGS 84');
    writeln(F,'Reserved 1');
    writeln(F,'Reserved 2');
    writeln(F,'Magnetic Variation,,,E');
    writeln(F,'Map Projection,Mercator,PolyCal,No,AutoCalOnly,No,BSBUseWPX,No');

    XMin:=round(Map.Width / 2 * (1 - Beta));
    XMax:=round(Map.Width / 2 * (1 + Beta));
    YMin:=round(Map.Height/ 2 * (1 - Beta));
    YMax:=round(Map.Height/ 2 * (1 + Beta));

    Pixel2WGS84(XMin,YMin,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('Point01,xy, %.4d, %.4d,in, deg,  %s, %s, grid,   ,           ,           ,N',
                                          [XMin, YMin, LatToStr(Lat), LongToStr(Long)]));
    Pixel2WGS84(XMax,YMin,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('Point02,xy, %.4d, %.4d,in, deg,  %s, %s, grid,   ,           ,           ,N',
                                          [XMax, YMin, LatToStr(Lat), LongToStr(Long)]));
    Pixel2WGS84(XMax,YMax,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('Point03,xy, %.4d, %.4d,in, deg,  %s, %s, grid,   ,           ,           ,N',
                                          [XMax, YMax, LatToStr(Lat), LongToStr(Long)]));
    Pixel2WGS84(XMin,YMax,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('Point04,xy, %.4d, %.4d,in, deg,  %s, %s, grid,   ,           ,           ,N',
                                          [XMin, YMax, LatToStr(Lat), LongToStr(Long)]));
    for I:=5 to 9 do
      writeln(F,Format('Point%.2d,xy,     ,     ,in, deg,    ,        ,N,    ,        ,E, grid,   ,           ,           ,N',[I]));
    for I:=10 to 30 do
      writeln(F,Format('Point%d,xy,     ,     ,in, deg,    ,        ,,    ,        ,, grid,   ,           ,           ,',[I]));

    writeln(F,'Map Feature = MF ; Map Comment = MC     These follow if they exist');
    writeln(F,'Track File = TF      These follow if they exist');
    writeln(F,'Moving Map Parameters = MM?    These follow if they exist');
    writeln(F,'MM0,Yes');
    writeln(F,'MMPNUM,4');
    writeln(F,'MMPXY,1,0,0');
    writeln(F,Format('MMPXY,2,%d,0',[Map.Width]));
    writeln(F,Format('MMPXY,3,%d,%d',[Map.Width, Map.Height]));
    writeln(F,Format('MMPXY,4,0,%d',[Map.Height]));
    Pixel2WGS84(0, 0, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,1,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Pixel2WGS84(Map.Width, 0, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,2,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Pixel2WGS84(Map.Width, Map.Height, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,3,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Pixel2WGS84(0, Map.Height, Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('MMPLL,4,  %d.%.6d,  %d.%.6d',
                              [trunc(Long),trunc(frac(abs(Long))*1000000),
                              trunc(Lat),trunc(frac(abs(Lat))*1000000)]));
    Scale:=0.5*(sqrt(sqr(Cal_Parameters[0, 0])+sqr(Cal_Parameters[0, 1]))+sqrt(sqr(Cal_Parameters[0, 3])+sqr(Cal_Parameters[0, 4])));
    writeln(F,Format('MM1B,%d.%.6d',[trunc(Scale),trunc(frac(Scale)*1000000)]));
    writeln(F,'MOP,Map Open Position,0,0');
    writeln(F,Format('IWH,Map Image Width/Height,%d,%d',[Map.Width, Map.Height]));

  finally
    CloseFile(F);
  end;

  // writing the calibration file
  // for TritonRMP (*.gmap)

  AssignFile(F,changeFileExt(FileName,'.gmap'));
  rewrite(F);
  OldDecimalSeparator:=DecimalSeparator;
  try
    DecimalSeparator:='.';

  // same XMin, Xmax, YMin, Ymax as for OziExplorer

    writeln(F,Format('Path=%s',[FileName]));
    writeln(F,Format('GBAname=%s',[copy(extractFileName(FileName),1,10)]));
    writeln(F,'Calib=0');

    Pixel2WGS84(XMin,YMin,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('%.4d %.4d %.8f %.8f 1',[XMin, YMin, Lat, Long]));

    Pixel2WGS84(XMax,YMin,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('%.4d %.4d %.8f %.8f 1',[XMax, YMin, Lat, Long]));

    Pixel2WGS84(XMax,YMax,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('%.4d %.4d %.8f %.8f 1',[XMax, YMax, Lat, Long]));

    Pixel2WGS84(XMin,YMax,Lat, Long);
    Lat:=Lat/Pi*180; Long:=Long/Pi*180;
    writeln(F,Format('%.4d %.4d %.8f %.8f 1',[Xmin, YMax, Lat, Long]));
    writeln(F,'0 0  0.00000000  0.00000000 0');
  finally
    DecimalSeparator:=OldDecimalSeparator;
    CloseFile(F);
  end;
end;

procedure TCalibrationProject.ExportPicLayer(FileName : string);
// create a calibration file for PicLayer Josm plugin
// http://josm.openstreetmap.de/wiki/Help/Plugin/PicLayer
var
  I : integer;
  XCenter, YCenter, IndiceSource : integer;
  LatCenter, LongCenter, LatH, LongH, LatV, LongV, degre, Scale : real;
  XCart, YCart, ZCart, XMerc, YMerc : real;
  F : TextFile;
  OldDecimalSeparator : char;
begin
  assert(Calculated,'Calibration not calculated in TCalibrationProject.ExportPicLayer');


  AssignFile(F,FileName);
  rewrite(F);
  try
    OldDecimalSeparator := DecimalSeparator;
    DecimalSeparator := '.';
    writeln(F, '#JOSM PicLayer plugin calibration data generated by Calibre_Carte');
    writeln(F, '# Date');

    XCenter := Map.Width div 2;
    YCenter := Map.Height div 2;
    Pixel2WGS84(XCenter,YCenter,LatCenter, LongCenter);
//    degre := 111900 * cos(LatCenter); // from PicLayer wiki
    degre := 111321 * cos(LatCenter);  // from GRS 80 a semi-major axe
//    degre := 110942 * cos(LatCenter);  // from GRS 80 b semi-minor axe
//    degre := 111195 * cos(LatCenter);  // from GRS 80 average radius (2*a + b)/3
    LatCenter := ln (tan (Pi/4 + LatCenter / 2));
    Pixel2WGS84(XCenter+100,YCenter,LatH, LongH);
    LatH := ln (tan (Pi/4 + LatH / 2));
    Pixel2WGS84(XCenter,YCenter-100,LatV, LongV);
    LatV := ln (tan (Pi/4 + LatV / 2));
    Scale := sqrt(sqr(LatH-LatCenter)+sqr(LongH-LongCenter));

    writeln(F, Format('SCALEY=%.15f',[sqrt(sqr(LatV-LatCenter)+sqr(LongV-LongCenter))/Scale]));
    writeln(F, Format('INITIAL_SCALE=%.15f',[Scale*Degre/Pi*180.0]));
    writeln(F, 'SCALEX=1.000000');

    IndiceSource:=0;
    donnees.VersGeocentriqueNum('WGS84', LatCenter, LongCenter, IndiceSource, XCart, YCart, Zcart);
    donnees.DepuisGeocentriqueNum(Xcart, YCart, Zcart, 'Mercator', IndiceSource, XMerc, YMerc);


{    writeln(F, Format('POSITION_Y=%.15f',[YMerc]));
    writeln(F, Format('POSITION_X=%.15f',[XMerc]));
    writeln(F, Format('INITIAL_POS_y=%.15f',[YMerc]));
    writeln(F, Format('INITIAL_POS_X=%.15f',[XMerc])); }

    writeln(F, Format('POSITION_Y=%.15f',[LatCenter*aWGS84]));
    writeln(F, Format('POSITION_X=%.15f',[LongCenter*aWGS84]));
    writeln(F, Format('INITIAL_POS_y=%.15f',[LatCenter*aWGS84]));
    writeln(F, Format('INITIAL_POS_X=%.15f',[LongCenter*aWGS84]));

//    writeln(F, Format('ANGLE=%.15f',[-arctan2(LatH-LatCenter, LongH-LongCenter)/Pi*180]));
//    writeln(F, Format('ANGLE=%.15f',[arctan2(LongV-LongCenter, LatV-LatCenter)/Pi*180]));
    writeln(F, Format('ANGLE=%.15f',[-arctan2((LatH-LatCenter)-(LongV-LongCenter), (LongH-LongCenter) + (LatV-LatCenter))/Pi*180]));
    writeln(F, 'SHEARX=0.0');
    writeln(F, 'SHEARY=0.0');

  finally
    DecimalSeparator:=OldDecimalSeparator;
    CloseFile(F);
  end;
end;


end.


