unit compute;

interface

type
  TRelation = record
    PixelX, PixelY : integer;
    CoordX, CoordY : real;
  end;

  TRelations = array of TRelation;

  TCal_Parameters = array [0..1] of array of real; // O for forward conversion, 1 for backward conversion

  procedure Compute_Quadratic(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
  procedure Compute_Homographic(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
  procedure Compute_Affine(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
  procedure Compute_Affine_svd(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
  procedure Compute_Affine_Isotrop(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
  procedure Compute_Affine_Isotrop_North(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
  procedure Compute_Affine_Isotrop_Scale(Relations : TRelations; Scale : real; var Cal_Parameters :TCal_Parameters);
  procedure Compute_Affine_Isotrop_North_Scale(Relations : TRelations; Scale : real; var Cal_Parameters :TCal_Parameters);


implementation

uses
  SysUtils, 
  Ap,
  Matrices, singular_value_decomposition;

// note : Ap is from http://www.alglib.net/


procedure Average(Relations : TRelations; var PixelXm, PixelYm, CoordXm, CoordYm : real);
var
  I : integer;
begin
  PixelXm:=0;
  PixelYm:=0;
  CoordXm:=0;
  CoordYm:=0;
  for I:=0 to length(Relations)-1 do
  begin
    PixelXm:=PixelXm+Relations[I].PixelX;
    PixelYm:=PixelYm+Relations[I].PixelY;
    CoordXm:=CoordXm+Relations[I].CoordX;
    CoordYm:=CoordYm+Relations[I].CoordY;
  end;
  PixelXm:=PixelXm/length(Relations);
  PixelYm:=PixelYm/length(Relations);
  CoordXm:=CoordXm/length(Relations);
  CoordYm:=CoordYm/length(Relations);
end;

procedure Compute_Quadratic(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
// U := a*x + b*y + c*x*y + d*x + e*y + f;
// V := g*x + h*y + i*x*y + j*x + k*y + l;
var
  A : TReal2DArray;
  X, B : TReal1DArray;

  I, N : integer;
begin
  assert(length(Relations)>=6,'Not enought reference points in Compute_Quadratic');
  assert(length(Cal_Parameters[0])=12,'Cal_Parameters not set to the right length in Compute_Quadratic');
  assert(length(Cal_Parameters[1])=12,'Cal_Parameters not set to the right length in Compute_Quadratic');

  N:=length(Relations);

  setlength(A, 2 * N, 12);
  setlength(X, 12);
  setlength(B, 2 * N);
  for I:=0 to N-1 do
  with Relations[I] do
  begin
    A[2*I, 0] := sqr(PixelX);
    A[2*I, 1] := sqr(PixelY);
    A[2*I, 2] := PixelX * PixelY;
    A[2*I, 3] := PixelX;
    A[2*I, 4] := PixelY;
    A[2*I, 5] := 1;
    A[2*I, 6] := 0;
    A[2*I, 7] := 0;
    A[2*I, 8] := 0;
    A[2*I, 9] := 0;
    A[2*I, 10] := 0;
    A[2*I, 11] := 0;
    A[2*I + 1, 0] := 0;
    A[2*I + 1, 1] := 0;
    A[2*I + 1, 2] := 0;
    A[2*I + 1, 3] := 0;
    A[2*I + 1, 4] := 0;
    A[2*I + 1, 5] := 0;
    A[2*I + 1, 6] := sqr(PixelX);
    A[2*I + 1, 7] := sqr(PixelY);
    A[2*I + 1, 8] := PixelX * PixelY;
    A[2*I + 1, 9] := PixelX;
    A[2*I + 1, 10] := PixelY;
    A[2*I + 1, 11] := 1;
    B[2*I] := CoordX;
    B[2*I + 1] := CoordY;
  end;

  SVD_Global(A, B, N, 12, X);

  for I:=0 to 11 do
    Cal_Parameters[0, I]:=X[I];

  // backward approximated by another quadratic transformation

  for I:=0 to N-1 do
  with Relations[I] do
  begin
    A[2*I, 0] := sqr(CoordX);
    A[2*I, 1] := sqr(CoordY);
    A[2*I, 2] := CoordX * CoordY;
    A[2*I, 3] := CoordX;
    A[2*I, 4] := CoordY;
    A[2*I, 5] := 1;
    A[2*I, 6] := 0;
    A[2*I, 7] := 0;
    A[2*I, 8] := 0;
    A[2*I, 9] := 0;
    A[2*I, 10] := 0;
    A[2*I, 11] := 0;
    A[2*I + 1, 0] := 0;
    A[2*I + 1, 1] := 0;
    A[2*I + 1, 2] := 0;
    A[2*I + 1, 3] := 0;
    A[2*I + 1, 4] := 0;
    A[2*I + 1, 5] := 0;
    A[2*I + 1, 6] := sqr(CoordX);
    A[2*I + 1, 7] := sqr(CoordY);
    A[2*I + 1, 8] := CoordX * CoordY;
    A[2*I + 1, 9] := CoordX;
    A[2*I + 1, 10] := CoordY;
    A[2*I + 1, 11] := 1;
    B[2*I] := PixelX;
    B[2*I + 1] := PixelY;
  end;

  SVD_Global(A, B, N, 12, X);

  for I:=0 to 11 do
    Cal_Parameters[1, I]:=X[I];

end;

procedure Compute_Homographic(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
// U := (a*x + b*y + c)/(g*x + h*y + 1);
// V := (d*x + e*y + f)/(g*x + h*y + 1);
var
  A : TReal2DArray;
  X, B : TReal1DArray;
  InvNorm : real;

  I, N : integer;
begin
  assert(length(Relations)>=3,'Not enought reference points in Compute_Homographic');
  assert(length(Cal_Parameters[0])=8,'Cal_Parameters not set to the right length in Compute_Homographic');
  assert(length(Cal_Parameters[1])=8,'Cal_Parameters not set to the right length in Compute_Homographic');

  N:=length(Relations);

  setlength(A, 2 * N, 8);
  setlength(X, 8);
  setlength(B, 2 * N);
  for I:=0 to N-1 do
  with Relations[I] do
  begin
    A[2*I, 0] := PixelX;
    A[2*I, 1] := PixelY;
    A[2*I, 2] := 1;
    A[2*I, 3] := 0;
    A[2*I, 4] := 0;
    A[2*I, 5] := 0;
    A[2*I, 6] := -PixelX * CoordX;
    A[2*I, 7] := -PixelY * CoordX;
    A[2*I + 1, 0] := 0;
    A[2*I + 1, 1] := 0;
    A[2*I + 1, 2] := 0;
    A[2*I + 1, 3] := PixelX;
    A[2*I + 1, 4] := PixelY;
    A[2*I + 1, 5] := 1;
    A[2*I + 1, 6] := -PixelX * CoordY;
    A[2*I + 1, 7] := -PixelY * CoordY;
    B[2*I] := CoordX;
    B[2*I + 1] := CoordY;
  end;

  SVD_Global(A, B, N, 8, X);

  for I:=0 to 7 do
    Cal_Parameters[0, I]:=X[I];

  // backward parameters calculation
  InvNorm := 1.0 / (Cal_Parameters[0, 0] * Cal_Parameters[0, 4] - Cal_Parameters[0, 1] * Cal_Parameters[0, 3]);
  Cal_Parameters[1, 0] := InvNorm * (Cal_Parameters[0, 4] * 1.0 - Cal_Parameters[0, 7] * Cal_Parameters[0, 5]);
  Cal_Parameters[1, 1] := InvNorm * (Cal_Parameters[0, 2] * Cal_Parameters[0, 7] - Cal_Parameters[0, 1] * 1.0);
  Cal_Parameters[1, 2] := InvNorm * (Cal_Parameters[0, 1] * Cal_Parameters[0, 5] - Cal_Parameters[0, 2] * Cal_Parameters[0, 4]);
  Cal_Parameters[1, 3] := InvNorm * (Cal_Parameters[0, 5] * Cal_Parameters[0, 6] - Cal_Parameters[0, 3] * 1.0);
  Cal_Parameters[1, 4] := InvNorm * (Cal_Parameters[0, 0] * 1.0 - Cal_Parameters[0, 2] * Cal_Parameters[0, 6]);
  Cal_Parameters[1, 5] := InvNorm * (Cal_Parameters[0, 2] * Cal_Parameters[0, 3] - Cal_Parameters[0, 0] * Cal_Parameters[0, 5]);
  Cal_Parameters[1, 6] := InvNorm * (Cal_Parameters[0, 3] * Cal_Parameters[0, 7] - Cal_Parameters[0, 4] * Cal_Parameters[0, 6]);
  Cal_Parameters[1, 7] := InvNorm * (Cal_Parameters[0, 1] * Cal_Parameters[0, 6] - Cal_Parameters[0, 0] * Cal_Parameters[0, 7]);
end;

procedure Compute_Affine_Backward(var Cal_Parameters :TCal_Parameters);
// use forward parameters to calculate backward one
var
  InvDet : real;
begin
  assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in Compute_Affine_Backward');
  assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in Compute_Affine_Backward');

  InvDet := 1.0 / (Cal_Parameters[0, 0] * Cal_Parameters[0, 4] - Cal_Parameters[0, 3] * Cal_Parameters[0, 1]);
  Cal_Parameters[1, 0] := Cal_Parameters[0, 4] * InvDet;
  Cal_Parameters[1, 1] := -Cal_Parameters[0, 1] * InvDet;
  Cal_Parameters[1, 2] := (Cal_Parameters[0, 1] * Cal_Parameters[0, 5] - Cal_Parameters[0, 2] * Cal_Parameters[0, 4]) * InvDet;
  Cal_Parameters[1, 3] := -Cal_Parameters[0, 3] * InvDet;
  Cal_Parameters[1, 4] := Cal_Parameters[0, 0] * InvDet;
  Cal_Parameters[1, 5] := -(Cal_Parameters[0, 0] * Cal_Parameters[0, 5] - Cal_Parameters[0, 2] * Cal_Parameters[0, 3]) * InvDet;
end;

procedure Compute_Affine(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
// U := a*x + b*y + c;
// V := d*x + e*y + f;
var
  I, N : integer;
  PixelXm, PixelYm, CoordXm, CoordYm : real;
  M11, M12, M21, M22, L1, L2, Det : real;
begin
  assert(length(Relations)>=3,'Not enought reference points in Compute_Affine');
  assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in Compute_Affine');
  assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in Compute_Affine');

  Average(Relations, PixelXm, PixelYm, CoordXm, CoordYm);
  N:=length(Relations);
  M11:=-n*sqr(PixelXm);
  M12:=-n*PixelXm*PixelYm;
  M21:=M12;
  M22:=-n*sqr(PixelYm);;
  L1:=-n*CoordXm*PixelXm;
  L2:=-n*CoordXm*PixelYm;;
  for I:=0 to N-1 do
   with Relations[I] do
  begin
    M11:=M11+sqr(PixelX);
    M12:=M12+PixelX*PixelY;
    M21:=M21+PixelX*PixelY;
    M22:=M22+sqr(PixelY);
    L1:=L1+PixelX*CoordX;
    L2:=L2+PixelY*CoordX;
  end;
  Det:=M11*M22-M12*M21;
  Cal_Parameters[0, 0]:=(L1*M22-L2*M12)/Det;
  Cal_Parameters[0, 1]:=(L2*M11-L1*M21)/Det;
  Cal_Parameters[0, 2]:=CoordXm - Cal_Parameters[0, 0]*PixelXm - Cal_Parameters[0, 1]*PixelYm;

  L1:=-n*CoordYm*PixelXm;
  L2:=-n*CoordYm*PixelYm;;
  for I:=0 to N-1 do
   with Relations[I] do
  begin
    L1:=L1+PixelX*CoordY;
    L2:=L2+PixelY*CoordY;
  end;
  Cal_Parameters[0, 3]:=(L1*M22-L2*M12)/Det;
  Cal_Parameters[0, 4]:=(L2*M11-L1*M21)/Det;
  Cal_Parameters[0, 5]:=CoordYm - Cal_Parameters[0, 3]*PixelXm - Cal_Parameters[0, 4]*PixelYm;

  Compute_Affine_Backward(Cal_Parameters);
end;

procedure Compute_Affine_svd(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
// U := a*x + b*y + c;
// V := d*x + e*y + f;
// same as Compute_Affine but with Singular Value Decomposition
var
  A : TReal2DArray;
  X, B : TReal1DArray;

  I, N : integer;
begin
  assert(length(Relations)>=3,'Not enought reference points in Compute_Affine_svd');
  assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in Compute_Affine_svd');
  assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in Compute_Affine_svd');

  N:=length(Relations);

  setlength(A, 2 * N, 6);
  setlength(X, 6);
  setlength(B, 2 * N);
  for I:=0 to N-1 do
  with Relations[I] do
  begin
    A[2*I, 0] := PixelX;
    A[2*I, 1] := PixelY;
    A[2*I, 2] := 1;
    A[2*I, 3] := 0;
    A[2*I, 4] := 0;
    A[2*I, 5] := 0;
    A[2*I + 1, 0] := 0;
    A[2*I + 1, 1] := 0;
    A[2*I + 1, 2] := 0;
    A[2*I + 1, 3] := PixelX;
    A[2*I + 1, 4] := PixelY;
    A[2*I + 1, 5] := 1;
    B[2*I] := CoordX;
    B[2*I + 1] := CoordY;
  end;

  SVD_Global(A, B, N, 6, X);

  for I:=0 to 5 do
    Cal_Parameters[0, I]:=X[I];

  Compute_Affine_Backward(Cal_Parameters);
end;


procedure Compute_Affine_Isotrop(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
var
  I, N : integer;
  PixelXm, PixelYm, CoordXm, CoordYm : real;
  Matrice4 : TMatrice4;
  Colonne4 : TColonne4;
begin
  assert(length(Relations)>=2,'Not enought reference points in Compute_Affine_Isotrop');
  assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop');
  assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop');

  // Isotropic =>
  // be carefull, Y axis are switched between pixels in picture and coordinates projection 
  // d := b;
  // e := -a;

  Average(Relations, PixelXm, PixelYm, CoordXm, CoordYm);
  Zero4(Matrice4, COlonne4);
  N:=length(Relations);

  Matrice4[3,1]:=N*PixelXm;
  Matrice4[4,1]:=-N*PixelYm;
  Matrice4[3,2]:=N*PixelYm;
  Matrice4[4,2]:=N*PixelXm;

  Matrice4[1,3]:=PixelXm;
  Matrice4[2,3]:=PixelYm;
  Matrice4[3,3]:=1;
  Colonne4[3]:=CoordXm;

  Matrice4[1,4]:=-PixelYm;
  Matrice4[2,4]:=PixelXm;
  Matrice4[4,4]:=1;
  Colonne4[4]:=CoordYm;

  for I:=0 to N-1 do
   with Relations[I] do
  begin
    Matrice4[1,1]:=Matrice4[1,1]+sqr(PixelX)+sqr(PixelY);
    Colonne4[1]:=Colonne4[1]+PixelX*CoordX-PixelY*CoordY;
    Colonne4[2]:=Colonne4[2]+PixelY*CoordX+PixelX*CoordY;
  end;
  Matrice4[2,2]:=Matrice4[1,1];

  Colonne4 := SolveLinear4(Matrice4, Colonne4);

  Cal_Parameters[0, 0]:=Colonne4[1];
  Cal_Parameters[0, 1]:=Colonne4[2];
  Cal_Parameters[0, 2]:=Colonne4[3];
  Cal_Parameters[0, 3]:=Colonne4[2];
  Cal_Parameters[0, 4]:=-Colonne4[1];
  Cal_Parameters[0, 5]:=Colonne4[4];

  Compute_Affine_Backward(Cal_Parameters);
end;

procedure Compute_Affine_Isotrop_North(Relations : TRelations; var Cal_Parameters :TCal_Parameters);
var
  I, N : integer;
  Numerator, Denominator : real;
  PixelXm, PixelYm, CoordXm, CoordYm : real;
begin
  assert(length(Relations)>=2,'Not enought reference points in Compute_Affine_Isotrop_North');
  assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop_North');
  assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop_North');

  Average(Relations, PixelXm, PixelYm, CoordXm, CoordYm);
  N:=length(Relations);
  Numerator:=N*(-CoordXm*PixelXm+CoordYm*PixelYm);
  Denominator:=-N*(sqr(PixelXm)+sqr(PixelYm));
  for I:=0 to N-1 do
   with Relations[I] do
  begin
    Numerator:=Numerator+PixelX*CoordX-PixelY*CoordY;
    Denominator:=Denominator+sqr(PixelX)+sqr(PixelY);
  end;

  Compute_Affine_Isotrop_North_Scale(Relations, Numerator/Denominator, Cal_Parameters);
end;

procedure Compute_Affine_Isotrop_Scale(Relations : TRelations; Scale : real; var Cal_Parameters :TCal_Parameters);
const
  MaxCycle = 1000;
var
  I, N, NbCycle : integer;
  Numerator, Denominator : real;
  Theta, Delta0 : real;
  cos0, sin0 : real;
  PixelXm, PixelYm, CoordXm, CoordYm : real;

  A : TReal2DArray;
  X, B : TReal1DArray;
begin
  assert(length(Relations)>=2,'Not enought reference points in Compute_Affine_Isotrop_Scale');
  assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop_Scale');
  assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop_Scale');
  assert(Scale<>0,'Scale equal to zero in Compute_Affine_Isotrop_Scale');

  // Isotropic =>
  // a := Scale * cos(Theta)
  // b := - Scale * sin(Theta)
  // d := - Scale * sin(Theta)
  // e := - Scale * cos(Theta)
  // we need to determied Theta (and also c and f)
  // we use root mean square for parameters determination
  // as it is not linear vs Theta (sin and cos), we do a local developement and
  // several iteration to obtain the right result

  Average(Relations, PixelXm, PixelYm, CoordXm, CoordYm);

  Theta:=0;
  NbCycle := 0;
  repeat
    inc(NbCycle);
    cos0:=cos(Theta);
    sin0:=sin(Theta);
{ the following code seems to be buggy
    Numerator:=0;
    Denominator:=0;
    for I:=0 to length(Relations)-1 do
     with Relations[I] do
    begin
      Numerator:=Numerator+
       (sin0*PixelX+cos0*PixelY)*
       (Scale*(cos0*(PixelX-PixelXm)-sin0*(PixelY-PixelYm))+CoordXm-CoordX)+
       (-cos0*PixelX+sin0*PixelY)*
       (Scale*(sin0*(PixelX-PixelXm)+cos0*(PixelY-PixelYm))-CoordYm+CoordY);
      Denominator:=Denominator+
       (sin0*PixelX+cos0*PixelY)*
       (sin0*(PixelX-PixelXm)+cos0*(PixelY-PixelYm))+
       (-cos0*PixelX+sin0*PixelY)*
       (cos0*(PixelX-PixelXm)-sin0*(PixelY-PixelYm));
    end;
    Delta0:=Numerator/Denominator; }

    N:=length(Relations);

    setlength(A, 2 * N, 3);
    setlength(X, 3);
    setlength(B, 2 * N);
    for I:=0 to N-1 do
    with Relations[I] do
    begin
      A[2*I, 0] := Scale * (- sin0 * PixelX - cos0 * PixelY);
      A[2*I, 1] := 1;
      A[2*I, 2] := 0;
      A[2*I + 1, 0] := Scale * (- cos0 * PixelX + sin0 * PixelY);
      A[2*I + 1, 1] := 0;
      A[2*I + 1, 2] := 1;
      B[2*I] := CoordX - scale * (cos0 * PixelX - sin0 * PixelY);
      B[2*I + 1] := CoordY + scale * (sin0 * PixelX + cos0 * PixelY);
    end;

    SVD_Global(A, B, N, 3, X);

{    for I:=0 to 5 do
      Cal_Parameters[0, I]:=X[I]; }
    Delta0 := X[0];


    Theta:=Theta + Delta0;
  until (abs(Delta0)<1e-10) or (NbCycle>=MaxCycle);

  if NbCycle>=MaxCycle then
    raise EZeroDivide.CreateFmt('Problem with Compute_Affine_Isotrop_Scale',[]);

  Cal_Parameters[0, 0] := Scale * cos(Theta);
  Cal_Parameters[0, 1] := - Scale * sin(Theta);
//  Cal_Parameters[0, 2] := CoordXm + scale * (cos(Theta) * PixelXm - sin(Theta) * PixelYm);
  Cal_Parameters[0, 2] := X[1];
  Cal_Parameters[0, 3] := - Scale * sin(Theta);
  Cal_Parameters[0, 4] := - Scale * cos(Theta);
//  Cal_Parameters[0, 5] := CoordYm - scale * (sin(Theta) * PixelXm + cos(Theta) * PixelYm);
  Cal_Parameters[0, 5] := X[2];

  Compute_Affine_Backward(Cal_Parameters);
end;

procedure Compute_Affine_Isotrop_North_Scale(Relations : TRelations; Scale : real; var Cal_Parameters :TCal_Parameters);
var
  PixelXm, PixelYm, CoordXm, CoordYm : real;
begin
  assert(length(Relations)>=1,'Not enought reference points in Compute_Affine_Isotrop_North_Scale');
  assert(length(Cal_Parameters[0])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop_North_Scale');
  assert(length(Cal_Parameters[1])=6,'Cal_Parameters not set to the right length in Compute_Affine_Isotrop_North_Scale');
  assert(Scale<>0,'Scale equal to zero in Compute_Affine_Isotrop_North_Scale');

  Average(Relations, PixelXm, PixelYm, CoordXm, CoordYm);

  Cal_Parameters[0, 0]:=Scale;
  Cal_Parameters[0, 1]:=0;
  Cal_Parameters[0, 2]:=CoordXm - Scale*PixelXm;
  Cal_Parameters[0, 3]:=0;
  Cal_Parameters[0, 4]:=-Scale;
  Cal_Parameters[0, 5]:=CoordYm + Scale*PixelYm;

  Compute_Affine_Backward(Cal_Parameters);
end;

end.
