unit control_dial;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls,
  carte, reference,
  donnee; // from Convertisseur

type
  TControlPointDial = class(TForm)
    ImageDetail: TImage;
    ButtonMoins: TButton;
    ButtonPlus: TButton;
    LabelScale: TLabel;
    Label1: TLabel;
    EditName: TEdit;
    GroupBox1: TGroupBox;
    Label2: TLabel;
    EditPixelX: TEdit;
    Label3: TLabel;
    EditPixelY: TEdit;
    GroupBox2: TGroupBox;
    LabelProjX: TLabel;
    LabelProjY: TLabel;
    EditProjX: TEdit;
    EditProjY: TEdit;
    ComboBoxProj: TComboBox;
    Label4: TLabel;
    EditProjZone: TEdit;
    ButtonGPX: TButton;
    ButtonOk: TButton;
    ButtonCancel: TButton;
    CheckBoxParis: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure ButtonPlusClick(Sender: TObject);
    procedure ButtonMoinsClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure EditPixelXChange(Sender: TObject);
    procedure ImageDetailMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ComboBoxProjChange(Sender: TObject);
    procedure ButtonOkClick(Sender: TObject);
    procedure ButtonGPXClick(Sender: TObject);
    procedure EditProjXKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure CheckBoxParisClick(Sender: TObject);
  private
    { Dclarations prives }
  public
    { Dclarations publiques }
    ScaleFactor : integer;
    Map : TImageCarte;
    NewPoint : boolean;
    procedure PrepareProjections;
    procedure PrepareNewPoint(const Mouse : TPoint);
    procedure PrepareOldPoint(const ThePoint : TReferencePoint);
    procedure Transfert2Point(const ThePoint : TReferencePoint);
    procedure DrawDetail;
    function Approximation(var Latitude, Longitude : real) : boolean;
    procedure ProposeApproximation;
  end;

var
  ControlPointDial: TControlPointDial;

implementation

{$R *.dfm}

uses
  Math,
  gpx_part, select_wpt, main,
  outils; // from Convertisseur


procedure TControlPointDial.FormCreate(Sender: TObject);
begin
  DecimalSeparator:='.';
  ScaleFactor:=5;
end;

procedure TControlPointDial.ButtonPlusClick(Sender: TObject);
begin
  if ScaleFactor=5 then
  begin
    ScaleFactor:=10;
    ButtonPlus.Enabled:=false;
  end
  else
  begin
    inc(ScaleFactor);
    ButtonMoins.Enabled:=true;
  end;
  LabelScale.Caption:=IntToStr(ScaleFactor)+':1';
  DrawDetail;
end;

procedure TControlPointDial.ButtonMoinsClick(Sender: TObject);
begin
  if ScaleFactor=10 then
  begin
    ScaleFactor:=5;
    ButtonPlus.Enabled:=true;
  end
  else
  begin
    dec(ScaleFactor);
    ButtonMoins.Enabled:=ScaleFactor<>1;
  end;
  LabelScale.Caption:=IntToStr(ScaleFactor)+':1';
  DrawDetail;
end;

procedure TControlPointDial.PrepareProjections;
var
  ListeProposition : TStringList;
  Ancien : string;
  I, Separations : integer;
begin
  ListeProposition:=TStringList.Create;
  Donnees.ListeSystemes(ListeProposition, Separations);
  with ComboBoxProj do
  begin
    Tag:=Separations; // on transfre dans le tag de la combobox la position des separations entre datums, proj. simples et proj. multiples
    if ItemIndex<>-1 then
      Ancien:=Items[ItemIndex]
    else
      Ancien:='';
    Items.Assign(ListeProposition);
    if Items.Count<>0 then
      ItemIndex:=0;
    for I:=0 to Items.Count-1 do
      if Ancien=Items[I] then
        ItemIndex:=I;
  end;
  ListeProposition.Free;
  ComboBoxProjChange(self);
end;

procedure TControlPointDial.PrepareNewPoint(const Mouse: TPoint);
begin
  EditName.Text:='';
  EditPixelX.Text:=IntToStr(Mouse.X);
  EditPixelY.Text:=IntToStr(Mouse.Y);
  EditProjX.Text:='';
  EditProjY.Text:='';
  EditProjZone.Text:='';
  NewPoint := true;
end;

procedure TControlPointDial.DrawDetail;
var
  Bitmap : TBitmap;
begin
  if Map<>nil then
  begin
    Bitmap := TBitmap.Create;
    Bitmap.Width:=ImageDetail.Width div ScaleFactor;
    Bitmap.Height:=ImageDetail.Height div ScaleFactor;
    Bitmap.PixelFormat:=pf24bit;
    try
      Map.Draw(Bitmap, StrToInt(EditPixelX.Text)-Bitmap.Width div 2,
                       StrToInt(EditPixelY.Text)-Bitmap.Height div 2,1);
      with ImageDetail.Canvas do
      begin
        StretchDraw(Rect(Point(0,0),Point(ImageDetail.Width-1,ImageDetail.Height-1)),Bitmap);
        Pen.Color:=clBlack;
        Pen.Width:=1;
        MoveTo(ImageDetail.Width div 2,0);
        LineTo(ImageDetail.Width div 2,ImageDetail.Height);
        MoveTo(0,ImageDetail.Height div 2);
        LineTo(ImageDetail.Width,ImageDetail.Height div 2);
      end;
    finally
      Bitmap.Free;
    end;
  end;
end;

procedure TControlPointDial.FormShow(Sender: TObject);
begin
  ButtonGPX.Enabled:=(GPXData.NbWayPoints>0);
  DrawDetail;
  EditName.SetFocus;
  ProposeApproximation;
end;

procedure TControlPointDial.EditPixelXChange(Sender: TObject);
begin
  if Visible then
    DrawDetail;
  ProposeApproximation;  
end;

procedure TControlPointDial.ImageDetailMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  try
    EditPixelX.Text:=IntToStr(StrToInt(EditPixelX.Text)+(X-ImageDetail.Width div 2) div ScaleFactor);
  except
  end;
  try
    EditPixelY.Text:=IntToStr(StrToInt(EditPixelY.Text)+(Y-ImageDetail.Height div 2) div ScaleFactor);
  except
  end;
  DrawDetail;
end;

procedure TControlPointDial.ComboBoxProjChange(Sender: TObject);
begin
//  EditDescription1.Text:=Donnees.LaDescription(ComboSource.Items[ComboSource.ItemIndex]);
  ComboBoxProj.Hint:=Donnees.LaDescription(ComboBoxProj.Items[ComboBoxProj.ItemIndex]);
  if ComboBoxProj.ItemIndex<Donnees.NbDatum then
  begin
    LabelProjX.Caption:='&Latitude';
    LabelProjY.Caption:='Lon&gitude';
    CheckBoxParis.Enabled:=true;
  end
  else begin
    LabelProjX.Caption:='&Easting (m)';
    LabelProjY.Caption:='&Northing (m)';
    CheckBoxParis.Enabled:=false;
  end;
  EditProjZone.Enabled:=
    (ComboBoxProj.ItemIndex>=Donnees.NbDatum+Donnees.NbProjection);
  Label4.Enabled:=EditProjZone.Enabled;

  ProposeApproximation;
end;

procedure TControlPointDial.ButtonOkClick(Sender: TObject);
// we try a conversion to WGS84 in order to check that datas are correct
var
  SX, SY : string;
  NumSource, NumDest : integer;

  procedure Stop(const Error_Msg : string);
  begin
    ShowMessage(Error_Msg);
    ModalResult:=mrNone;
  end;

begin
  try
    try
      StrToInt(EditPixelX.Text);
    except
      EditPixelX.SetFocus;
      Raise EConvertError.Create('La valeur en X du pixel est incorrecte, exemple 1285 (sans virgule)');
    end;

    try
      StrToInt(EditPixelY.Text);
    except
      EditPixelY.SetFocus;
      Raise EConvertError.Create('La valeur en Y du pixel est incorrecte, exemple 1285 (sans virgule)');
    end;

    if ComboBoxProj.ItemIndex<Donnees.NbDatum then
    begin
      if CheckBoxParis.Checked then
        NumSource:=1
      else
        NumSource:=0;
    end
    else begin
      if EditProjZone.Enabled then
      begin
        try
          NumSource:=StrToInt(EditProjZone.Text);
        except
          EditProjZone.SetFocus;
          Raise EConvertError.Create('Le numro de fuseau est incorrecte, exemple 12');
        end;
      end
      else
        NumSource:=0;
    end;

    NumDest:=0;

    Donnees.Conversion(ComboBoxProj.Text, 'WGS84',
                       EditProjX.Text, EditProjY.Text,
                       NumSource, NumDest,
                       SX, SY, uaDegre);
  except
    on E: Exception do Stop(E.Message);
  end;
end;

procedure TControlPointDial.Transfert2Point(const ThePoint: TReferencePoint);
begin
  ThePoint.Name:=EditName.Text;
  ThePoint.XPixel:=StrToInt(EditPixelX.Text);
  ThePoint.YPixel:=StrToInt(EditPixelY.Text);
  ThePoint.Num_Proj:=ComboBoxProj.ItemIndex;
  ThePoint.Latitude:=EditProjX.Text;
  ThePoint.Longitude:=EditProjY.Text;
  if ComboBoxProj.ItemIndex>=Donnees.NbDatum+Donnees.NbProjection then
    ThePoint.Zone:=StrToInt(EditProjZone.Text)
  else
    ThePoint.Zone:=0;
  ThePoint.Paris:=CheckBoxParis.Checked;
end;

procedure TControlPointDial.PrepareOldPoint(const ThePoint: TReferencePoint);
begin
  NewPoint := false;
  EditName.Text:=ThePoint.Name;
  EditPixelX.Text:=IntToStr(ThePoint.XPixel);
  EditPixelY.Text:=IntToStr(ThePoint.YPixel);
  ComboBoxProj.ItemIndex:=ThePoint.Num_Proj;
  EditProjX.Text:=ThePoint.Latitude;
  EditProjY.Text:=ThePoint.Longitude;
  if ComboBoxProj.ItemIndex>=Donnees.NbDatum+Donnees.NbProjection then
    EditProjZone.Text:=IntToStr(ThePoint.Zone)
  else
    EditProjZone.Text:='';
  CheckBoxParis.Checked:=ThePoint.Paris;

  ComboBoxProjChange(self);
end;

procedure TControlPointDial.ButtonGPXClick(Sender: TObject);
var
  ALat, ALong : real;
begin
  if Approximation(ALat, ALong) then
    WaypointDlg.Prepare(true, ALat, ALong)
  else
    WaypointDlg.Prepare(false, 0, 0);
  if WaypointDlg.ShowModal=mrOk then
  begin
    NewPoint :=false;
    ComboBoxProj.ItemIndex:=0;
    EditProjX.Text:=FloatToStrF(WaypointDlg.GPXLat, ffFixed, 10, 7)+'';
    EditProjY.Text:=FloatToStrF(WaypointDlg.GPXLong, ffFixed, 11, 7)+'';
    EditName.Text:=WaypointDlg.ComboBoxWpt.Text;
  end;
end;

function TControlPointDial.Approximation(var Latitude, Longitude : real) : boolean;
// calculate the approximative coordinates (WGS84) of the pixel position
// return True if succesfull, False if not
begin
  if (Form1.CalibrationProject<>nil) and Form1.CalibrationProject.Calculated then
  begin
    try
      Form1.CalibrationProject.Pixel2WGS84(StrToInt(EditPixelX.Text), StrToInt(EditPixelY.Text), Latitude, Longitude);
      Result := true;
    except
      Result := false;
    end;
  end
  else
    Result :=false;
end;

procedure TControlPointDial.ProposeApproximation;
// using approximative coordinates of the pixel, try to propose coordinates with rounding
const
  Tolerance = 10; // tolerance distance in pixel
  Eps = 1e-5;
var
  Lat, Long : real;
  X, Y, XPlus, YPlus : real;
  Zone, ZonePlus : integer;
  Delta, Delta_Deg, Delta_Gr, Delta_Min, Delta_Sec : real;
  Grad_level, Grad_Level_Deg, Grad_Level_Gr, Grad_Level_Min, Grad_Level_Sec : integer; // graduation level, =0 in [1,2], =1 in [2,5], =2 in [5,10], =3 in [10,20] ...
  Best_Level, Which_Best : integer;
  XFrac, YFrac : integer;

  function Get_Grad_Size (N : integer) : real;
  begin
    result := 0;
    if N>=0 then
    begin
      case N mod 3 of
        0 : result := IntPower(10.0, N div 3);
        1 : result := 2.0 * IntPower(10.0, N div 3);
        2 : result := 5.0 * IntPower(10.0, N div 3);
      end;
    end
    else
    begin
      case N mod 3 of
        0 : result := IntPower(10.0, N div 3);
        -1 : result := 0.5 * IntPower(10.0, N div 3);
        -2 : result := 0.2 * IntPower(10.0, N div 3);
      end;
    end;
  end;

begin
  if NewPoint then
  begin
    if Approximation(Lat, Long) then
    begin
      if CheckBoxParis.Checked then
        Zone := 1
      else
        Zone := 0;
      ZonePlus := Zone;
      Donnees.ConversionNum('WGS84', ComboBoxProj.Text, Lat, Long, 0, Zone, X, Y);

      // we are looking for a rounding at less than Tolerance pixels
      try
        Form1.CalibrationProject.Pixel2WGS84(StrToInt(EditPixelX.Text)+Tolerance, StrToInt(EditPixelY.Text), Lat, Long);
        Donnees.ConversionNum('WGS84', ComboBoxProj.Text, Lat, Long, 0, ZonePlus, XPlus, YPlus);
        if ZonePlus<>Zone then
          raise Exception.Create('Point on a zone limit');
        Delta := max(abs(XPlus-X), abs(YPlus-Y));

        if ComboBoxProj.ItemIndex<Donnees.NbDatum then
        begin  // datum
        // looking if rounding is better in , grade, minute or second

          Grad_Level_Deg := Floor(log10(Delta/Pi*180)) * 3;
          while Get_Grad_Size(Grad_Level_Deg+1)<Delta/Pi*180 do
            inc(Grad_Level_Deg);
          Delta_Deg := Get_Grad_Size(Grad_Level_Deg);
          XPlus := round(X/Pi*180/Delta_Deg)*Delta_Deg;
          YPlus := round(Y/Pi*180/Delta_Deg)*Delta_Deg;
          while (abs(XPlus/Get_Grad_Size(Grad_Level_Deg+1)-round(XPlus/Get_Grad_Size(Grad_Level_Deg+1)))<Eps) and
                (abs(YPlus/Get_Grad_Size(Grad_Level_Deg+1)-round(YPlus/Get_Grad_Size(Grad_Level_Deg+1)))<Eps) and
                (Grad_Level_Deg<7) do  // stopping  360
            inc(Grad_Level_Deg);
          Best_Level := Grad_Level_Deg;
          Which_Best := 0;

          Grad_Level_Gr := Floor(log10(Delta/Pi*200)) * 3;
          while Get_Grad_Size(Grad_Level_Gr+1)<Delta/Pi*200 do
            inc(Grad_Level_Gr);
          Delta_Gr := Get_Grad_Size(Grad_Level_Gr);
          XPlus := round(X/Pi*200/Delta_Gr)*Delta_Gr;
          YPlus := round(Y/Pi*200/Delta_Gr)*Delta_Gr;
          while (abs(XPlus/Get_Grad_Size(Grad_Level_Gr+1)-round(XPlus/Get_Grad_Size(Grad_Level_Gr+1)))<Eps) and
                (abs(YPlus/Get_Grad_Size(Grad_Level_Gr+1)-round(YPlus/Get_Grad_Size(Grad_Level_Gr+1)))<Eps) and
                (Grad_Level_Gr<7) do  // stopping  400 g 
            inc(Grad_Level_Gr);
          if Grad_Level_Gr>Best_Level then
          begin
            Best_Level := Grad_Level_Gr;
            Which_Best := 1;
          end;


          Grad_Level_Min := Floor(log10(Delta/Pi*180*60)) * 3;
          while Get_Grad_Size(Grad_Level_Min+1)<Delta/Pi*180*60 do
            inc(Grad_Level_Min);
          Delta_Min := Get_Grad_Size(Grad_Level_Min);
          XPlus := round(X/Pi*180*60/Delta_Min)*Delta_Min;
          YPlus := round(Y/Pi*180*60/Delta_Min)*Delta_Min;
          if Grad_Level_Min>4 then
            Grad_Level_Min :=4
          else
            while (abs(XPlus/Get_Grad_Size(Grad_Level_Min+1)-round(XPlus/Get_Grad_Size(Grad_Level_Min+1)))<Eps) and
                  (abs(YPlus/Get_Grad_Size(Grad_Level_Min+1)-round(YPlus/Get_Grad_Size(Grad_Level_Min+1)))<Eps) and
                  (Grad_Level_Min<4) do
              inc(Grad_Level_Min);
          if Grad_Level_Min>Best_Level+5 then
          begin
            Best_Level := Grad_Level_Min-5;
            Which_Best := 2;
          end;

          Grad_Level_Sec := Floor(log10(Delta/Pi*180*3600)) * 3;
          while Get_Grad_Size(Grad_Level_Sec+1)<Delta/Pi*180*3600 do
            inc(Grad_Level_Sec);
          Delta_Sec := Get_Grad_Size(Grad_Level_Sec);
          XPlus := round(X/Pi*180*3600/Delta_Sec)*Delta_Sec;
          YPlus := round(Y/Pi*180*3600/Delta_Sec)*Delta_Sec;
          if Grad_Level_Sec>4 then
            Grad_Level_Sec :=4
          else
            while (abs(XPlus/Get_Grad_Size(Grad_Level_Sec+1)-round(XPlus/Get_Grad_Size(Grad_Level_Sec+1)))<Eps) and
                  (abs(YPlus/Get_Grad_Size(Grad_Level_Sec+1)-round(YPlus/Get_Grad_Size(Grad_Level_Sec+1)))<Eps) and
                  (Grad_Level_Sec<4) do
              inc(Grad_Level_Sec);
          if Grad_Level_Sec>Best_Level+10 then
          begin
            Best_Level := Grad_Level_Sec-10;
            Which_Best := 3;
          end;

          Delta := Get_Grad_Size(Best_Level);

          case Which_Best of
            0 : begin  // degree
              XPlus := round(X/Pi*180/Delta)*Delta;
              YPlus := round(Y/Pi*180/Delta)*Delta;
              if Grad_Level_Deg>=0 then
              begin
                EditProjX.Text := Format('%d', [round(XPlus)]);
                EditProjY.Text := Format('%d', [round(YPlus)]);
              end
              else
              begin
                XFrac := round(frac(XPlus)*IntPower(10, -(Grad_Level_Deg+1) div 3 +1));
                YFrac := round(frac(YPlus)*IntPower(10, -(Grad_Level_Deg+1) div 3 +1));
                EditProjX.Text := Format('%d.%d', [trunc(XPlus+Eps*sign(XPlus)), abs(XFrac)]);
                EditProjY.Text := Format('%d.%d', [trunc(YPlus+Eps*sign(YPlus)), abs(YFrac)]);
              end;
            end;
            1 : begin  // grade
              XPlus := round(X/Pi*200/Delta)*Delta;
              YPlus := round(Y/Pi*200/Delta)*Delta;
              if Grad_Level_Gr>=0 then
              begin
                EditProjX.Text := Format('%dg', [round(XPlus)]);
                EditProjY.Text := Format('%dg', [round(YPlus)]);
              end
              else
              begin
                XFrac := round(frac(XPlus)*IntPower(10, -(Grad_Level_Gr+1) div 3 +1));
                YFrac := round(frac(YPlus)*IntPower(10, -(Grad_Level_Gr+1) div 3 +1));
                EditProjX.Text := Format('%d.%dg', [trunc(XPlus+Eps*sign(XPlus)), abs(XFrac)]);
                EditProjY.Text := Format('%d.%dg', [trunc(YPlus+Eps*sign(YPlus)), abs(YFrac)]);
              end;
            end;
            2 : begin  // minut
              XPlus := round(X/Pi*180*60/Delta)*Delta;
              YPlus := round(Y/Pi*180*60/Delta)*Delta;
              if Grad_Level_Min>=0 then
              begin
                EditProjX.Text := Format('%d%d''', [round(XPlus) div 60, round(abs(XPlus)) mod 60]);
                EditProjY.Text := Format('%d%d''', [round(YPlus) div 60, round(abs(YPlus)) mod 60]);
              end
              else
              begin
                XFrac := round(frac(XPlus)*IntPower(10, -(Grad_Level_Min+1) div 3 +1));
                YFrac := round(frac(YPlus)*IntPower(10, -(Grad_Level_Min+1) div 3 +1));
                EditProjX.Text := Format('%d%d.%d''', [trunc(XPlus+Eps*sign(XPlus)) div 60, trunc(abs(XPlus)+Eps) mod 60, abs(XFrac)]);
                EditProjY.Text := Format('%d%d.%d''', [trunc(XPlus+Eps*sign(YPlus)) div 60, trunc(abs(YPlus)+Eps) mod 60, abs(YFrac)]);
              end;
            end;
            3 : begin  // Second
              XPlus := round(X/Pi*180*3600/Delta)*Delta;
              YPlus := round(Y/Pi*180*3600/Delta)*Delta;
              if Grad_Level_Sec>=0 then
              begin
                EditProjX.Text := Format('%d%d''%d''''', [round(XPlus) div 3600, (round(abs(XPlus)) div 60) mod 60, round(abs(XPlus)) mod 60]);
                EditProjY.Text := Format('%d%d''%d''''', [round(YPlus) div 3600, (round(abs(YPlus)) div 60) mod 60, round(abs(YPlus)) mod 60]);
              end
              else
              begin
                XFrac := round(frac(XPlus)*IntPower(10, -(Grad_Level_Sec+1) div 3 +1));
                YFrac := round(frac(YPlus)*IntPower(10, -(Grad_Level_Sec+1) div 3 +1));
                EditProjX.Text := Format('%d%d''%d.%d''''', [trunc(XPlus+Eps*sign(XPlus)) div 3600, (trunc(abs(XPlus)+Eps) div 60) mod 60, trunc(abs(XPlus)+Eps) mod 60, abs(XFrac)]);
                EditProjY.Text := Format('%d%d''%d.%d''''', [trunc(YPlus+Eps*sign(YPlus)) div 3600, (trunc(abs(YPlus)+Eps) div 60) mod 60, trunc(abs(YPlus)+Eps) mod 60, abs(YFrac)]);
              end;
            end;
          end;
        end
        else
        begin // projection (single or multiple)
          Grad_Level := Floor(log10(Delta)) * 3;
          while Get_Grad_Size(Grad_Level+1)<Delta do
            inc(Grad_Level);
          Delta := Get_Grad_Size(Grad_Level);

          XPlus := round(X/Delta)*Delta;
          YPlus := round(Y/Delta)*Delta;

          if (abs(XPlus/(Delta*10)-round(XPlus/(Delta*10)))>Eps) or
             (abs(YPlus/(Delta*10)-round(YPlus/(Delta*10)))>Eps)
          then // not a nice rounding, less than one order of magnitude, compared to the delta area
            raise Exception.Create('Not a real rounding');

          EditProjX.Text := IntToStr(round(XPlus));
          EditProjY.Text := IntToStr(round(YPlus));
          if ComboBoxProj.ItemIndex>=Donnees.NbDatum+Donnees.NbProjection then
            EditProjZone.Text := IntToStr(Zone);
        end;



      except
        if ComboBoxProj.ItemIndex<Donnees.NbDatum then
        begin  // datum
          EditProjX.Text := Format('%.6f', [X/Pi*180]);
          EditProjY.Text := Format('%.6f', [Y/Pi*180]);
        end
        else
        begin // projection (single or multiple)
          EditProjX.Text := IntToStr(round(X));
          EditProjY.Text := IntToStr(round(Y));
          if ComboBoxProj.ItemIndex>=Donnees.NbDatum+Donnees.NbProjection then
            EditProjZone.Text := IntToStr(Zone);
        end;
      end;
    end;
  end;
end;

procedure TControlPointDial.EditProjXKeyDown(Sender: TObject;
  var Key: Word; Shift: TShiftState);
begin
  NewPoint := false;
end;

procedure TControlPointDial.CheckBoxParisClick(Sender: TObject);
begin
  ProposeApproximation;
end;

end.
