unit reference;

interface

type

  TProjectionType = (ptDatum,ptProjection,ptMultiple);

  TReferencePoint = class
  private
    FYPixel: integer;
    FNum_Proj: integer;
    FXPixel: integer;
    FLongitude: string;
    FLatitude: string;
    FZone: integer;
    FName: string;
    FParis: boolean;
    FErrorDistance: real;
    procedure SetLatitude(const Value: string);
    procedure SetLongitude(const Value: string);
    procedure SetNum_Proj(const Value: integer);
    procedure SetXPixel(const Value: integer);
    procedure SetYPixel(const Value: integer);
    procedure SetZone(const Value: integer);
    procedure SetName(const Value: string);
    procedure SetParis(const Value: boolean);
    procedure SetErrorDistance(const Value: real);
    function GetName_Proj : string;
    function GetProjectionType : TProjectionType;
  public
    property Name : string read FName write SetName;
    property XPixel : integer read FXPixel write SetXPixel;
    property YPixel : integer read FYPixel write SetYPixel;
    property Num_Proj : integer read FNum_Proj write SetNum_Proj;
    property Name_Proj : string read GetName_Proj;
    property ProjectionType : TProjectionType read GetProjectionType;
    property Latitude : string read FLatitude write SetLatitude;
    property Longitude : string read FLongitude write SetLongitude;
    property Zone : integer read FZone write SetZone; // only for multi-projection
    property Paris : boolean read FParis write SetParis; // only for Datum, indicate if orgine is in Paris instead of Greenwich
    property ErrorDistance : real read FErrorDistance write SetErrorDistance;
    constructor Create; virtual;
    destructor Destroy; override;
    procedure LoadFromString(S : string);
    function SaveToString : string;
    procedure Assign(Source : TReferencePoint);
  end;

implementation

{ TReferencePoint }

uses
  donnee, datum, ancetre, // from Convertisseur
  SysUtils, Math;

constructor TReferencePoint.Create;
begin
  ErrorDistance:=NaN;
end;

destructor TReferencePoint.Destroy;
begin

  inherited;
end;

procedure TReferencePoint.LoadFromString(S: string);

  function GetNextString(var Source : string) : string;
  begin
    delete(Source,1, pos('"',Source)); // find and remove first "
    result:=copy(S,1,pos('"',Source)-1);
    delete(Source,1, pos('"',Source)); // find and remove second "
  end;
  function GetNextInteger(var Source : string) : integer;
  begin
    while Source[1]=' ' do
      delete(Source,1,1); // remonving spaces before the number
    if pos(' ',Source)=0 then
    begin // number is the last element of the line
      result:=StrToInt(Source);
      Source:='';
    end
    else
    begin // number is NOT the last element of the line
      result:=StrToInt(copy(Source,1,pos(' ',Source)-1));
      delete(Source,1,pos(' ',Source));
    end;
  end;

var
  NewName : string;
begin
  Name:=GetNextString(S);
  XPixel:=GetNextInteger(S);
  YPixel:=GetNextInteger(S);

  NewName:=GetNextString(S);

  Num_Proj:=Donnees.NumeroDatum(NewName);
  if Num_Proj<>-1 then
  begin // datum
    Latitude:=GetNextString(S);
    Longitude:=GetNextString(S);
    while S[1]=' ' do
      delete(S,1,1); // remonving spaces before the boolean
    Paris:=UpperCase(S)='TRUE';
  end
  else
  begin // projections
    Num_Proj:=Donnees.NumeroProjection(NewName);
    if Num_Proj<>-1 then
    begin // projection simple
      Num_Proj:=Num_Proj+Donnees.NbDatum;
      Latitude:=GetNextString(S);
      Longitude:=GetNextString(S);
    end
    else
    begin  // projection multiple
      Num_Proj:=Donnees.NumeroMultiple(NewName);
      Num_Proj:=Num_Proj+Donnees.NbDatum+Donnees.NbProjection;
      Latitude:=GetNextString(S);
      Longitude:=GetNextString(S);
      Zone:=GetNextInteger(S);
    end;
  end;
end;

function TReferencePoint.SaveToString: string;
begin
  if Num_Proj<Donnees.NbDatum then
  begin // datum
    result:=Format('"%s" %d %d "%s" "%s" "%s"',
     [Name, XPixel, YPixel, (Donnees.LesDatum[Num_Proj] as TDatum).Nom, Latitude, Longitude]);
    if Paris then
      result:=result+' TRUE'
    else
      result:=result+' FALSE';
  end
  else
  begin
    if Num_Proj<Donnees.NbDatum+Donnees.NbProjection then
    begin // projection simple
      result:=Format('"%s" %d %d "%s" "%s" "%s"',
    [Name, XPixel, YPixel, (Donnees.LesProjection[Num_Proj-Donnees.NbDatum] as TAncetre).Nom, Latitude, Longitude]);
    end
    else
    begin // projection multiple
      result:=Format('"%s" %d %d "%s" "%s" "%s" %d',
    [Name, XPixel, YPixel, (Donnees.LesMultiple[Num_Proj-Donnees.NbDatum-Donnees.NbProjection] as TAncetre).Nom,
                   Latitude, Longitude, Zone]);
    end
  end;
end;

procedure TReferencePoint.SetErrorDistance(const Value: real);
begin
  FErrorDistance := Value;
end;

procedure TReferencePoint.SetLatitude(const Value: string);
begin
  FLatitude := Value;
end;

procedure TReferencePoint.SetLongitude(const Value: string);
begin
  FLongitude := Value;
end;

procedure TReferencePoint.SetName(const Value: string);
begin
  FName := Value;
end;

procedure TReferencePoint.SetNum_Proj(const Value: integer);
begin
  FNum_Proj := Value;
end;

procedure TReferencePoint.SetParis(const Value: boolean);
begin
  FParis := Value;
end;

procedure TReferencePoint.SetXPixel(const Value: integer);
begin
  FXPixel := Value;
end;

procedure TReferencePoint.SetYPixel(const Value: integer);
begin
  FYPixel := Value;
end;

procedure TReferencePoint.SetZone(const Value: integer);
begin
  FZone := Value;
end;

function TReferencePoint.GetName_Proj : string;
begin
  case ProjectionType of
    ptDatum : result:=(Donnees.LesDatum[Num_Proj] as TDatum).Nom;
    ptProjection : result:=(Donnees.LesProjection[Num_Proj-Donnees.NbDatum] as TAncetre).Nom;
    ptMultiple : result:=(Donnees.LesMultiple[Num_Proj-Donnees.NbDatum-Donnees.NbProjection] as TAncetre).Nom;
  end;
end;

function TReferencePoint.GetProjectionType : TProjectionType;
begin
  if Num_Proj<Donnees.NbDatum then
    result:=ptDatum
  else
  begin
    if Num_Proj<Donnees.NbDatum+Donnees.NbProjection then
      result:=ptProjection
    else
      result:=ptMultiple;
  end;
end;

procedure TReferencePoint.Assign(Source : TReferencePoint);
begin
  FYPixel := Source.YPixel;
  FNum_Proj := Source.Num_Proj;
  FXPixel := Source.XPixel;
  FLongitude := Source.Longitude;
  FLatitude := Source.Latitude;
  FZone := Source.Zone;
  FName := Source.Name;
  FParis := Source.Paris;
end;

end.
