unit select_wpt;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls;

type
  TWaypointDlg = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    Bevel1: TBevel;
    ComboBoxWpt: TComboBox;
    EditCommentaire: TEdit;
    EditDescription: TEdit;
    procedure ComboBoxWptChange(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    Correspondance : array of integer;
    GPXLat, GPXLong : real;
    procedure Prepare(Approx : boolean; LatApprox, LongApprox : real);
  end;

var
  WaypointDlg: TWaypointDlg;

implementation

{$R *.dfm}

uses
  gpx_part;

{ TWaypointDlg }

procedure TWaypointDlg.Prepare(Approx : boolean; LatApprox, LongApprox : real);
// prepare the list of waypoints
// if approximative coordinates are known, order waypoints using distance to the approximation
var
  I, J, K, Previous : integer;
  Distances : array of real;
  D : real;
begin
  assert(GPXData.NbWayPoints<>0,'No GPX waypoints in TWaypointDlg.Prepare');
  setlength(Correspondance, GPXData.NbWayPoints);
  if Approx then
  begin
    ComboBoxWpt.Items.Clear;
    setlength(Distances, GPXData.NbWayPoints);
    for I:=0 to GPXData.NbWayPoints-1 do
    with GPXData.WayPoints[I] do
    begin
      // calculating the distance, indeed scalar product of unit vector from earth center
      // i.e. D = cos(Alpha), high D mind closer distance
      D := cos(Latitude/180*Pi) * cos(Longitude/180*Pi) * cos(LatApprox) * cos(LongApprox) +
           cos(Latitude/180*Pi) * sin(Longitude/180*Pi) * cos(LatApprox) * sin(LongApprox) +
           sin(Latitude/180*Pi)                         * sin(LatApprox);
      J :=0;
      while (J<I) and (D<Distances[J]) do
        inc(J);
      ComboBoxWpt.Items.Insert(J, Utf8ToAnsi(GPXData.WayPoints[I].Name));
      for K:=I downto J+1 do
      begin
        Distances[K] := Distances[K-1];
        Correspondance[K] := Correspondance[K-1];
      end;
      Distances[J] := D;
      Correspondance[J] := I;

    end;
    ComboBoxWpt.ItemIndex:=0;

  end
  else
  begin
    Previous := ComboBoxWpt.ItemIndex;
    ComboBoxWpt.Items.Clear;
    if Previous=-1 then
      Previous:=0;
    for I:=0 to GPXData.NbWayPoints-1 do
    begin
      ComboBoxWpt.Items.Add(Utf8ToAnsi(GPXData.WayPoints[I].Name));
      Correspondance[I] := I;
    end;
    ComboBoxWpt.ItemIndex:=Previous;
  end;

  ComboBoxWptChange(self);
end;

procedure TWaypointDlg.ComboBoxWptChange(Sender: TObject);
begin
  with GPXData.WayPoints[Correspondance[ComboBoxWpt.ItemIndex]] do
  begin
    EditCommentaire.Text:=Utf8ToAnsi(Comment);
    EditDescription.Text:=Utf8ToAnsi(Description);
    GPXLat := Latitude;
    GPXLong := Longitude;
  end;  
end;

end.
