unit singular_value_decomposition;

interface

uses
  Ap;

procedure SVD_Global(A : TReal2DArray; B : TReal1DArray; N, M : integer; var X : TReal1DArray);
// prepare SVD calculation and use results back
// A . X = B
// N is the number of reference points
// M is the number of unknown parameters
// A is the matrix to be decomposited : size 2N x M
// B is the vector to be combined with the decomposition : size 2N
// X is the result vector : size M


implementation

uses
  Math, SysUtils,
  svd;

// note : SVD is from http://www.alglib.net/  


procedure SVD_Global(A : TReal2DArray; B : TReal1DArray; N, M : integer; var X : TReal1DArray);
// prepare SVD calculation and use results back
// A . X = B
// N is the number of reference points
// M is the number of unknown parameters
// A is the matrix to be decomposited : size 2N x M
// B is the vector to be combined with the decomposition : size 2N
// X is the result vector : size M
var
  U, VT : TReal2DArray;
  W : TReal1DArray;
  I, J : integer;
{$IFDEF DEBUG}
  Old_A, New_A : TReal2DArray;
  K : integer;
{$ENDIF}
begin
  assert(high(B)=2*N-1,'Wrong size of B in SVD_Global');
  assert(high(X)=M-1,'Wrong size of X in SVD_Global');
  assert(high(A)=2*N-1,'Wrong height of A in SVD_Global');
  assert(high(A[0])=M-1,'Wrong width of A in SVD_Global');

  setlength(VT, M, M);
  setlength(U, 2*N, M);
  setlength(W, M);

{$IFDEF DEBUG}
  setlength(Old_A, 2*N, M);
  setlength(New_A, 2*N, M);
  for I:=0 to 2*N -1 do
    for J:=0 to M - 1 do
      Old_A[I, J]:=A[I, J];
{$ENDIF}

// warning svdcmp used transposed matrix i.e. A[Y, X], not A[X, Y]

  if not RMatrixSVD(A, 2*N, M, 2, 2, 2, W, U, VT) then
    raise EZeroDivide.CreateFmt('Problem with RMatrixSVD',[]);


{$IFDEF DEBUG}
//  checking A := U W VT
  for I:=0 to 2*N -1 do
  begin
    for J:=0 to M - 1 do
    begin
      New_A[I, J]:=0;
      for K:=0 to M-1 do
        New_A[I, J] := New_A[I, J] + U[I,K] * W[K] * VT[K, J];
    end;
  end;

{$ENDIF}

//  X := V * W * Ut * b

// first V := V* W
  for I:=0 to M - 1 do
  begin
    if abs(W[I])<MachineEpsilon then
      raise EZeroDivide.CreateFmt('SVD is singular',[]);
    for J:=0 to M - 1 do
      VT[I, J] := VT[I, J] / W [I];
  end;

// second W := transpose(U) * B
  for I:=0 to M-1 do
  begin
    W[I] := 0;
    for J:=0 to 2 * N -1 do
      W[I] := W[I] + U[J, I] * B[J];
  end;

// third X := (V * W ) * (tU * B) i.e. V * W
  for I:=0 to M - 1 do
  begin
    X[I]:= 0;
    for J:=0 to M - 1 do
      X[I] := X[I] + VT[J, I] * W[J];
  end;

end;

end.
